# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base class for JS Callbacks
"""
import hashlib
import logging

from py_mini_racer.py_mini_racer import (  # type: ignore
    JSEvalException,
    JSTimeoutException,
)

from ..binding_accessor import BindingAccessor
from ..constants import LIFECYCLE_METHODS
from ..exceptions import SqreenException
from ..rules import RuleCallback
from ..utils import (
    ALL_STRING_CLASS,
    CustomJSONEncoder,
    create_bound_method,
    to_unicode_safe,
)
from .mini_racer_pool import JSContextPool

LOGGER = logging.getLogger(__name__)


class JSException(SqreenException):
    """ Base exception raised in JSCB
    """

    def __init__(self, message, callback, arguments):
        super(JSException, self).__init__(message)
        self.callback = callback
        self.arguments = arguments

    def exception_infos(self):
        return {"cb": self.callback, "args": self.arguments}


class JSCB(RuleCallback):
    """ A callback that run a JS function as pre / post / failing through
    py_mini_racer context.
    """

    SUPPORTS_BUDGET = True

    def __init__(self, *args, **kwargs):
        # Create pre / post / failing callback methods on the instance early on
        # because RuleCallback.__init__ needs them.
        rule_callbacks = frozenset((kwargs.get("callbacks") or {}).keys())
        for callback_name in LIFECYCLE_METHODS.values():
            if callback_name in rule_callbacks:
                self.create_callback(callback_name)

        super(JSCB, self).__init__(*args, **kwargs)

        self._loaded = False
        self.arguments = {}
        self.source = ''
        self.code_id = None

    def create_callback(self, name):
        """ Create an execute callback alias called `name`.
        """
        def callback(self, instance, args, kwargs, **options):
            return self.execute(name, instance, args, kwargs, **options)

        try:
            callback.__name__ = name
            callback.__qualname__ = name
        except AttributeError:
            pass
        method = create_bound_method(callback, self)
        setattr(self, name, method)

    def load(self):
        """ Create the binding accessors and save the js definitions
        """

        self.source = ''
        if self.callbacks:
            for callback_name, callback_args in self.callbacks.items():

                if not isinstance(callback_args, list):
                    self.source += "{} = {}\n".format(callback_name, callback_args)
                    self.arguments[callback_name] = []
                else:
                    self.source += "{} = {}\n".format(callback_name, callback_args[-1])
                    arguments = callback_args[:-1]
                    self.arguments[callback_name] = [
                        BindingAccessor(arg) for arg in arguments
                    ]

        self.code_id = hashlib.md5(self.source.encode('utf-8')).hexdigest()
        self._loaded = True

    def execute(self, name, instance, args, kwargs, **options):
        """ Fetches a context before executing _execute
        """
        if self._loaded is False:
            self.load()

        budget_in_ms = self.get_remaining_budget(options)
        if budget_in_ms is not None and budget_in_ms <= 0:
            LOGGER.debug("Skipping execution of callback %s because budget is negative", name)
            return

        pool = JSContextPool.fetch_for_runner(self.runner)
        return pool.with_context(
            lambda ctx: self._execute(
                ctx, name, budget_in_ms or 0, self.arguments[name],
                instance, args, kwargs, **options))

    def _execute(self, js_ctx, name, budget_in_ms, arguments,
                 instance, args, kwargs, result=None, exc_info=None,
                 **options):
        """ Execute a JS callback passed in definition.
        Handle recording attack, observations and chaining.
        Protected against infinite recursion with a max number of JS calls
        set to 100.
        """

        if self.code_id in js_ctx.failed_code_ids:
            LOGGER.debug("Skipping execution of callback %s (code md5 %s)"
                         " due to previous failure of definition eval",
                         name, self.code_id)
            return

        if self.code_id not in js_ctx.code_ids:
            js_ctx.add_code(self.code_id, self.source)

        request = self.storage.get_current_request()
        response = self.storage.get_current_response()

        # Fallback on a blank request for binding accessor
        if request is None:
            request = self.storage.default_request_class(storage=self.storage)

        return_value = result if name == "post" else exc_info

        # Safeguard against infinite recursion
        for _ in range(100):
            binding_eval_args = {
                "binding": locals(),
                "global_binding": globals(),
                "request": request,
                "response": response,
                "inst": instance,
                "args": self.storage.get_current_args(args),
                "kwargs": kwargs,
                "data": self.data,
                "rv": return_value,
            }

            resolved_args = [
                arg.resolve(**binding_eval_args) for arg in arguments
            ]
            if name in self.conditions:
                resolved_args = self._restrict(name, resolved_args)

            LOGGER.debug("Resolved args %s for %s", resolved_args, arguments)

            try:
                callee = "sqreen_data['{}']['{}']".format(self.code_id, name)
                result = js_ctx.call(
                    callee, *resolved_args,
                    encoder=CustomJSONEncoder, timeout=int(budget_in_ms)
                )
            except JSTimeoutException as err:
                LOGGER.debug("Budget exhausted while running JS callback: %s", err)
                return None
            except JSEvalException as err:
                raise JSException(err.args[0], name, resolved_args)

            LOGGER.debug("JS Result %r for %s", result, self.rule_name)

            if result is None:
                return result

            # Process the return value
            self._record_attack(result)
            self._record_observations(result)

            # Check for chaining
            if result.get("call") is None:
                return result

            # Prepare next call
            name = result["call"]

            if name not in self.callbacks:
                raise JSException(
                    "Invalid callback '{}'".format(name), name, None
                )

            return_value = result.get("data")

            if result.get("args"):
                arguments = [BindingAccessor(arg) for arg in result["args"]]
            else:
                arguments = self.arguments[name]

    def _record_attack(self, return_value):
        """ Record an attack if the JS callback returned a record info
        """
        if return_value.get("record"):
            self.record_attack(return_value["record"])

    def _record_observations(self, return_value):
        """ Record observations if the JS callback returned a observations list
        """
        if return_value.get("observations"):
            for observation in return_value["observations"]:
                self.record_observation(*observation)

    def _restrict(self, name, arguments):
        """ Filter out useless values from iterables present in *arguments*.
        It only supports conditions with a single hash_val_include.
        """
        if name not in self.conditions or name not in self.arguments:
            return arguments
        condition = self.conditions[name]
        expressions = [
            accessor.expression for accessor in self.arguments[name]
        ]
        for value, iterable, min_length in self.iter_hash_val_include_values(
            condition
        ):
            try:
                value_idx = expressions.index(value)
                iterable_idx = expressions.index(iterable)
            except ValueError:
                continue
            arguments[iterable_idx] = self.hash_value_included(
                arguments[value_idx], arguments[iterable_idx], int(min_length)
            )
        return arguments

    @classmethod
    def iter_hash_val_include_values(cls, condition):
        """ Yield arguments of operator %hash_val_includes in *condition*.
        """
        values = condition.get("%hash_val_include")
        if len(condition) == 1 and values is not None:
            yield values

    @classmethod
    def hash_value_included(cls, needed, iterable, min_length=8, max_depth=20):
        """ Return a filtered, deep copy of dict *iterable*, where only
        (sub)values included in *needed* are present.
        """
        needed = to_unicode_safe(needed)
        result = {}
        insert = []
        todos = [(result, key, value, 0) for key, value in iterable.items()]
        while todos:
            where, key, value, depth = todos.pop()
            if not isinstance(key, (int, ALL_STRING_CLASS)):
                key = str(key)
            if depth >= max_depth:
                insert.append((where, key, value))
            elif isinstance(value, dict):
                val = {}
                insert.append((where, key, val))
                todos.extend((val, k, v, depth + 1) for k, v in value.items())
            elif isinstance(value, list):
                val = []
                insert.append((where, key, val))
                todos.extend(
                    (val, k, v, depth + 1) for k, v in enumerate(value)
                )
            elif value is not None and needed is not None:
                v = to_unicode_safe(value)
                if len(v) < min_length or v not in needed:
                    pass
                elif isinstance(where, list):
                    where.append(value)
                else:
                    where[key] = value
        for where, key, value in reversed(insert):
            if not value:
                pass
            elif isinstance(where, list):
                where.append(value)
            else:
                where[key] = value
        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base framework strategy
"""
import logging

from .import_hook import BaseStrategy

LOGGER = logging.getLogger(__name__)


class HookpointMismatchError(Exception):
    """ Raised when a callback is trying to be added for a strategy and their
    hookpoint are not the same
    """

    pass


class FrameworkStrategy(BaseStrategy):
    """ Specific strategy for framework instrumentation. They hook on atrributes
    defined in the class, MODULE_NAME, HOOK_CLASS, HOOK_METHOD. They wrap the
    resulting object with self.wrapper by passing original as first argument and
    a middleware for the correct framework as second argument
    """

    def __init__(
        self,
        strategy_key,
        observation_queue,
        queue,
        import_hook,
        before_hook_point=None,
    ):
        super(FrameworkStrategy, self).__init__(
            observation_queue, queue, import_hook, before_hook_point
        )
        self.strategy_key = strategy_key

        # These values should be defined by subclasses
        self.middleware = None
        self.wrapper = None

    def add_callback(self, callback):
        """ Add a callback on a specific endpoint defined by the couple
        (hook_module, hook_name) embedded in callback
        """

        # Check that callback hookpoint and strategy hookpoint match
        strategy_module = "{}::{}".format(self.MODULE_NAME, self.HOOK_CLASS)
        mismatch_module = callback.hook_module != strategy_module
        mismatch_method = callback.hook_name != self.HOOK_METHOD
        if mismatch_module or mismatch_method:
            err_msg = "Callback hookpoint ({}, {}) doesn't match strategy callback ({}, {})"
            msg = err_msg.format(
                callback.hook_module,
                callback.hook_name,
                strategy_module,
                self.HOOK_METHOD,
            )
            raise HookpointMismatchError(msg)

        super(FrameworkStrategy, self).add_callback(callback)

    def hook(self):
        """
        Once hooked, the middleware will call the callbacks at the right moment.
        """

        # Check if we already hooked at
        if not self.hooked:
            self.import_hook.register_patcher(
                self.MODULE_NAME,
                self.HOOK_CLASS,
                self.HOOK_METHOD,
                self.import_hook_callback,
            )

            self.hooked = True

    def import_hook_callback(self, original):
        """ Monkey-patch the object located at hook_class.hook_name on an
        already loaded module
        """
        return self.wrapper(original, self.middleware)

    @classmethod
    def get_strategy_id(cls, callback):
        """ This strategy only hook on
        (cls.MODULE_NAME::cls.HOOK_CLASS, cls.HOOK_METHOD)
        """
        return (
            "{}::{}".format(cls.MODULE_NAME, cls.HOOK_CLASS),
            cls.HOOK_METHOD,
        )

    def _restore(self):
        """ The hooked module will always stay hooked
        """
        pass



#
# This file is part of pyasn1 software.
#
# Copyright (c) 2005-2018, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/pyasn1/license.html
#
try:
    from collections import OrderedDict

except ImportError:
    OrderedDict = dict

from ... import debug
from ... import error
from ...type import base
from ...type import char
from ...type import tag
from ...type import univ
from ...type import useful

__all__ = ['encode']


class AbstractItemEncoder(object):
    def encode(self, value, encodeFun, **options):
        raise error.PyAsn1Error('Not implemented')


class BooleanEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return bool(value)


class IntegerEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return int(value)


class BitStringEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return str(value)


class OctetStringEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return value.asOctets()


class TextStringEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return str(value)


class NullEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return None


class ObjectIdentifierEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return str(value)


class RealEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return float(value)


class SetEncoder(AbstractItemEncoder):
    protoDict = dict

    def encode(self, value, encodeFun, **options):
        value.verifySizeSpec()

        namedTypes = value.componentType
        substrate = self.protoDict()

        for idx, (key, subValue) in enumerate(value.items()):
            if namedTypes and namedTypes[idx].isOptional and not value[idx].isValue:
                continue
            substrate[key] = encodeFun(subValue, **options)
        return substrate


class SequenceEncoder(SetEncoder):
    protoDict = OrderedDict


class SequenceOfEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        value.verifySizeSpec()
        return [encodeFun(x, **options) for x in value]


class ChoiceEncoder(SequenceEncoder):
    pass


class AnyEncoder(AbstractItemEncoder):
    def encode(self, value, encodeFun, **options):
        return value.asOctets()


tagMap = {
    univ.Boolean.tagSet: BooleanEncoder(),
    univ.Integer.tagSet: IntegerEncoder(),
    univ.BitString.tagSet: BitStringEncoder(),
    univ.OctetString.tagSet: OctetStringEncoder(),
    univ.Null.tagSet: NullEncoder(),
    univ.ObjectIdentifier.tagSet: ObjectIdentifierEncoder(),
    univ.Enumerated.tagSet: IntegerEncoder(),
    univ.Real.tagSet: RealEncoder(),
    # Sequence & Set have same tags as SequenceOf & SetOf
    univ.SequenceOf.tagSet: SequenceOfEncoder(),
    univ.SetOf.tagSet: SequenceOfEncoder(),
    univ.Choice.tagSet: ChoiceEncoder(),
    # character string types
    char.UTF8String.tagSet: TextStringEncoder(),
    char.NumericString.tagSet: TextStringEncoder(),
    char.PrintableString.tagSet: TextStringEncoder(),
    char.TeletexString.tagSet: TextStringEncoder(),
    char.VideotexString.tagSet: TextStringEncoder(),
    char.IA5String.tagSet: TextStringEncoder(),
    char.GraphicString.tagSet: TextStringEncoder(),
    char.VisibleString.tagSet: TextStringEncoder(),
    char.GeneralString.tagSet: TextStringEncoder(),
    char.UniversalString.tagSet: TextStringEncoder(),
    char.BMPString.tagSet: TextStringEncoder(),
    # useful types
    useful.ObjectDescriptor.tagSet: OctetStringEncoder(),
    useful.GeneralizedTime.tagSet: OctetStringEncoder(),
    useful.UTCTime.tagSet: OctetStringEncoder()
}

# Type-to-codec map for ambiguous ASN.1 types
typeMap = {
    univ.Set.typeId: SetEncoder(),
    univ.SetOf.typeId: SequenceOfEncoder(),
    univ.Sequence.typeId: SequenceEncoder(),
    univ.SequenceOf.typeId: SequenceOfEncoder(),
    univ.Choice.typeId: ChoiceEncoder(),
    univ.Any.typeId: AnyEncoder()
}


class Encoder(object):

    # noinspection PyDefaultArgument
    def __init__(self, tagMap, typeMap={}):
        self.__tagMap = tagMap
        self.__typeMap = typeMap

    def __call__(self, value, **options):
        if not isinstance(value, base.Asn1Item):
            raise error.PyAsn1Error('value is not valid (should be an instance of an ASN.1 Item)')

        if debug.logger & debug.flagEncoder:
            logger = debug.logger
        else:
            logger = None

        if logger:
            debug.scope.push(type(value).__name__)
            logger('encoder called for type %s <%s>' % (type(value).__name__, value.prettyPrint()))

        tagSet = value.tagSet

        try:
            concreteEncoder = self.__typeMap[value.typeId]

        except KeyError:
            # use base type for codec lookup to recover untagged types
            baseTagSet = tag.TagSet(value.tagSet.baseTag, value.tagSet.baseTag)

            try:
                concreteEncoder = self.__tagMap[baseTagSet]

            except KeyError:
                raise error.PyAsn1Error('No encoder for %s' % (value,))

        if logger:
            logger('using value codec %s chosen by %s' % (concreteEncoder.__class__.__name__, tagSet))

        pyObject = concreteEncoder.encode(value, self, **options)

        if logger:
            logger('encoder %s produced: %s' % (type(concreteEncoder).__name__, repr(pyObject)))
            debug.scope.pop()

        return pyObject


#: Turns ASN.1 object into a Python built-in type object(s).
#:
#: Takes any ASN.1 object (e.g. :py:class:`~pyasn1.type.base.PyAsn1Item` derivative)
#: walks all its components recursively and produces a Python built-in type or a tree
#: of those.
#:
#: One exception is that instead of :py:class:`dict`, the :py:class:`OrderedDict`
#: can be produced (whenever available) to preserve ordering of the components
#: in ASN.1 SEQUENCE.
#:
#: Parameters
#: ----------
#  asn1Value: any pyasn1 object (e.g. :py:class:`~pyasn1.type.base.PyAsn1Item` derivative)
#:     pyasn1 object to encode (or a tree of them)
#:
#: Returns
#: -------
#: : :py:class:`object`
#:     Python built-in type instance (or a tree of them)
#:
#: Raises
#: ------
#: :py:class:`~pyasn1.error.PyAsn1Error`
#:     On encoding errors
#:
#: Examples
#: --------
#: Encode ASN.1 value object into native Python types
#:
#: .. code-block:: pycon
#:
#:    >>> seq = SequenceOf(componentType=Integer())
#:    >>> seq.extend([1, 2, 3])
#:    >>> encode(seq)
#:    [1, 2, 3]
#:
encode = Encoder(tagMap, typeMap)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Patch an async event loop class with a context-aware task factory."""

import logging

from ..._vendors.wrapt import FunctionWrapper
from .import_hook import ImportHookStrategy

LOGGER = logging.getLogger(__name__)


def _patch_loop_cls(loop_cls):
    """Patch an async event loop class with a context-aware task factory.

    Return the modified class constructor.
    """
    orig_init = loop_cls.__init__
    orig_set_task_factory = loop_cls.set_task_factory

    def __init__wrapper(wrapped, instance, args, kwargs):
        wrapped(*args, **kwargs)
        instance.set_task_factory(None)

    def set_task_factory_wrapper(wrapped, instance, args, kwargs):
        from ...async_context import create_task_factory

        new_args = list(args)
        new_args[0] = create_task_factory(args[0])
        wrapped(*new_args, **kwargs)

    setattr(loop_cls, "__init__",
            FunctionWrapper(orig_init, __init__wrapper))
    setattr(loop_cls, "set_task_factory",
            FunctionWrapper(orig_set_task_factory, set_task_factory_wrapper))
    return loop_cls


class AsyncEventLoopStrategy(ImportHookStrategy):
    """Patch an async event loop class with a context-aware task factory."""

    def import_hook_callback(self, original):
        return _patch_loop_cls(original)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Enforce security responses for SDK users or IP addresses.
"""

import logging

from ..actions import ActionName
from ..binding_accessor import BindingAccessor
from ..rules import RuleCallback
from ..sdk import events

LOGGER = logging.getLogger(__name__)


class SecurityResponsesIP(RuleCallback):

    INTERRUPTIBLE = False

    def pre(self, instance, args, kwargs, **options):
        request = self.storage.get_current_request()
        if request is None or self.runner is None:
            return
        client_ip = request.raw_client_ip
        if client_ip is None:
            return
        action = self.runner.action_store.get_for_ip(client_ip)
        if action is None:
            return
        events.track_action(action, {"ip_address": str(client_ip)},
                            storage=self.storage)
        if action.name == ActionName.BLOCK_IP:
            LOGGER.debug(
                "IP %s is blacklisted by action %s",
                client_ip,
                action.iden,
            )
            return {
                "status": "action_block",
                "action_id": action.iden,
                "immediate": True,
            }
        elif action.name == ActionName.REDIRECT_IP:
            LOGGER.debug(
                "IP %s is redirected to %r by action %s",
                client_ip,
                action.target_url,
                action.iden,
            )
            return {
                "status": "action_redirect",
                "action_id": action.iden,
                "target_url": action.target_url,
                "immediate": True,
            }


# Compatibility alias
IPActionCB = SecurityResponsesIP


class SecurityResponsesUser(RuleCallback):

    def __init__(self, *args, **kwargs):
        super(SecurityResponsesUser, self).__init__(*args, **kwargs)
        arguments = self.callbacks.get("pre")
        if arguments is not None:
            self._arguments = [BindingAccessor(ba) for ba in arguments]
        else:
            self._arguments = None

    def pre(self, instance, args, kwargs, **options):
        """Trigger an action for a user when required."""
        if self.runner is None:
            return

        request = self.storage.get_current_request()
        response = self.storage.get_current_response()

        binding_eval_args = {
            "request": request,
            "response": response,
            "inst": instance,
            "args": self.storage.get_current_args(args),
            "kwargs": kwargs,
            "data": self.data,
        }

        resolved_args = binding_eval_args["args"]
        if self._arguments is not None:
            resolved_args = [
                arg.resolve(**binding_eval_args) for arg in self._arguments]

        if not resolved_args:
            return

        user_dict = resolved_args[0]
        action = self.runner.action_store.get_for_user(user_dict)
        if action is None:  # No action planned for this user, do nothing.
            return

        events.track_action(action, {"user": user_dict}, storage=self.storage)

        if action.name == ActionName.BLOCK_USER:
            LOGGER.debug("User %r is blocked by action %r", user_dict, action.iden)
            return {
                "status": "action_block",
                "action_id": action.iden,
                "immediate": True,
            }
        elif action.name == ActionName.REDIRECT_USER:
            LOGGER.debug(
                "User %r is redirected to %r by action %r",
                user_dict,
                action.target_url,
                action.iden,
            )
            return {
                "status": "action_redirect",
                "action_id": action.iden,
                "target_url": action.target_url,
                "immediate": True,
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Signal Client API
"""
import logging

from ._vendors.sqreen_security_signal_sdk import Client
from .config import CONFIG
from .http_client import USER_AGENT

LOGGER = logging.getLogger(__name__)


class SignalClient(Client):

    user_agent = USER_AGENT


def get_signal_client(session, url=None, batch_size=0, max_staleness=0,
                      user_agent=None):
    """ Get a Signal Client.
    """
    if user_agent is not None:
        SignalClient.user_agent = user_agent
    client = SignalClient(
        token=session.session_token,
        max_batch_size=batch_size,
        interval_batch=max_staleness,
        session_token=True,
        proxy_url=CONFIG["PROXY_URL"],
        base_url=url or CONFIG["INGESTION_URL"],
    )
    return client



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" WSGI Middleware
"""
import logging
import sys

from .base import BaseMiddleware

LOGGER = logging.getLogger(__name__)


class WSGIApplicationWrapper:
    """
    WSGI Wrapper injected by the middleware to wrap the WSGI application.
    The pre callbacks are called with the WSGI environ and can override it.
    The post callbacks are called after the original application without arguments.
    The failing callbacks are called when the original application has failed.
    """

    def __init__(self, middleware, app, environ, start_response, *args):
        self.middleware = middleware
        self.orig_app = app
        self.orig_start_response = start_response
        self.orig_args = args
        self.orig_write_func = None
        try:
            new_args = middleware.execute_pre_callbacks((environ,))
            if new_args:
                environ = new_args[0]
            self.it = self.serve_application(environ)
        except Exception as e:
            resp = middleware.execute_failing_callbacks(e, args=(environ,))
            if resp is None:
                raise
            # override the response
            self._start_response(resp.status, list(resp.headers.items()),
                                 sys.exc_info())
            self.it = [resp.body or b""]
            self.late_post = False
        else:
            # If we don't have the original write function, start_response
            # was not called yet and will probably be called by the iterator.
            self.late_post = self.orig_write_func is None
            if not self.late_post:
                # XXX it is not possible to override the response here because
                # we can't call start_response again without an exception. It
                # is thus not possible to alter the response (without changing
                # the Content-Length header).
                middleware.execute_post_callbacks(None, args=(environ,),
                                                  record_attack=False)

        if self.it:
            self.orig_close = getattr(self.it, "close", None)
            self.orig_len = getattr(self.it, "__len__", None)
            if self.orig_len:
                self.__len__ = self.orig_len

    def _start_response(self, *args):
        self.orig_write_func = self.start_response(*args)
        return self.write

    def serve_application(self, environ):
        """
        Wrapper around the original WSGI application.
        """
        return self.orig_app(environ, self._start_response, *self.orig_args)

    def start_response(self, status, headers, *args):
        """
        Wrapper around the original WSGI start_response function.
        """
        return self.orig_start_response(status, headers, *args)

    def write(self, data):
        """
        Wrapper around the original WSGI write function returned by
        start_response.
        If you plan to hook this function, it's probably better to use iterate.
        """
        return self.orig_write_func(data)

    def iterate(self, data):
        """
        Wrapper around the iterator returned by the WSGI application.
        It is called on each entry yielded by the iterator.
        """
        return data

    def close(self):
        """
        Wrapper around the original close function on the iterator returned by
        the WSGI application.
        """
        orig_close = getattr(self, "orig_close", None)
        if orig_close is not None:
            self.orig_close()

    def __iter__(self):
        try:
            for data in self.it:
                yield self.iterate(data)
        except Exception as e:
            if not self.orig_write_func:
                # start_response was not called yet, or was called during
                # the first iteration and failed
                resp = self.middleware.execute_failing_callbacks(e)
                if resp is None:
                    raise
                self._start_response(resp.status, list(resp.headers.items()),
                                     sys.exc_info())
                yield resp.body or b""
        else:
            if self.late_post:
                # Late post does not have access to the environ as it might have
                # been freed already
                self.middleware.execute_post_callbacks(None, record_attack=False)
        finally:
            self.close()


class WSGIMiddleware(BaseMiddleware):

    def wrap_app(self, *args):
        return WSGIApplicationWrapper(self, *args)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
from datetime import datetime
from enum import Enum
from typing import Any, List, Sequence, TypedDict, Union


class SignalType(str, Enum):
    POINT = "point"
    METRIC = "metric"


class SignalProperties(TypedDict, total=False):
    actor: Any
    context: Any
    context_schema: str
    location: Any
    location_infra: Any
    payload_schema: str
    source: str
    trigger: Any
    time: datetime
    type: SignalType


class Signal(SignalProperties):
    signal_name: str
    payload: Any


class TraceProperties(SignalProperties, total=False):
    signal_name: str
    payload: Any


class Trace(TraceProperties):
    data: Sequence[Signal]


AnySignal = Union[Signal, Trace]


class Batch(List[AnySignal]):
    pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Sqreen exceptions
"""

import logging
import traceback

LOGGER = logging.getLogger(__name__)


class SqreenException(Exception):
    """ Base exception for all sqreen exceptions
    """

    def __str__(self):
        return self.__repr__()

    def exception_infos(self):
        return {}


class InvalidApplicationName(SqreenException):
    """ Exception raised when the application name is invalid.
    """
    pass


class InvalidSignature(SqreenException):
    """ Exception raised when encountering an invalid rule signature
    """

    def __init__(self, name, rulespack_id):
        self.callback_payload = {
            "rule_name": name,
            "rulespack_id": rulespack_id,
        }


class InvalidArgument(SqreenException):
    """ Exception raised when sqreen code receive invalid arguments like bad
    rule dict.
    """

    pass


class RequestBlocked(SqreenException):
    """Base class for blocking requests (because of attack or action)."""

    def __init__(self):
        self._pending = True

    def done(self):
        """Mark the blocking as done.

        If the exception is not marked as done, an error is triggered when
        the object is destroyed.
        """
        self._pending = False

    def __del__(self):
        if not self._pending:
            return  # All right.
        LOGGER.error(
            "Exception %r was destroyed but it is pending!",
            self.__class__.__name__,
        )
        LOGGER.debug("\n".join(traceback.format_stack()))


# This exception name is particularly important since it is often seen by
# Sqreen users when watching their logs. It should not raise any concern to
# them.
class AttackBlocked(RequestBlocked):
    """ Raised when a callback detected an attack
    """

    def __init__(self, rule_name):
        super(AttackBlocked, self).__init__()
        self.rule_name = rule_name

    def __repr__(self):
        msg = "Sqreen blocked a security threat (type: #{}). No action is required."
        return msg.format(self.rule_name)


class BaseAction(RequestBlocked):
    """Base class when a callback triggers a security action."""

    def __init__(self, action_id):
        super(BaseAction, self).__init__()
        self.action_id = action_id


class ActionBlock(BaseAction):
    """Exception raised when a security action blocks a request."""

    def __repr__(self):
        return (
            "Sqreen blocked a request (action_id: {}). "
            "No action is required.".format(self.action_id)
        )


class ActionRedirect(BaseAction):
    """Exception raised when a security action redirects a request."""

    def __init__(self, action_id, target_url):
        super(ActionRedirect, self).__init__(action_id)
        self.target_url = target_url

    def __repr__(self):
        return (
            "Sqreen redirected a request to {!r} (action_id: {}). "
            "No action is required".format(self.target_url, self.action_id)
        )


class MissingDataException(Exception):
    """ Exception that occurs when trying to access undefined binding accessor """

    def __init__(self, binding_accessor):
        self.binding_accessor = binding_accessor


###
# HTTP Exceptions
###


class SqreenHttpException(SqreenException):
    pass


class InvalidResponseContentType(SqreenHttpException):
    def __init__(self, content_type):
        self.content_type = content_type

    def __repr__(self):
        return "Invalid response Content-Type: {!r}".format(self.content_type)


class InvalidJsonResponse(SqreenHttpException):
    def __init__(self, parsing_exception):
        self.parsing_exception = parsing_exception

    def __repr__(self):
        msg = "An error occurred while trying to parse the response: {!r}"
        return msg.format(self.parsing_exception)


class StatusFailedResponse(SqreenHttpException):
    def __init__(self, response):
        self.response = response

    def __repr__(self):
        msg = "Response returned with a status false: {!r}"
        return msg.format(self.response)


class InvalidStatusCodeResponse(SqreenHttpException):
    def __init__(self, status, response_data=None):
        self.status = status
        self.response_data = response_data

    def __repr__(self):
        msg = "Response status code is invalid: {!r}"
        return msg.format(self.status)


# Incompotible exceptions that may be raised at startup


class UnsupportedFrameworkVersion(Exception):
    def __init__(self, framework, version):
        self.framework = framework
        self.version = version

    def __str__(self):
        return "{} in version {} is not supported".format(
            self.framework.title(), self.version
        )


class UnsupportedPythonVersion(Exception):
    def __init__(self, python_version):
        self.python_version = python_version

    def __str__(self):
        return "Python version {} is unsupported".format(self.python_version)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base middleware for peripheric frameworks integration
"""
import logging

from ...constants import ACTIONS
from ...exceptions import ActionBlock, ActionRedirect, AttackBlocked
from ..hook_point import (
    execute_failing_callbacks,
    execute_post_callbacks,
    execute_pre_callbacks,
)

LOGGER = logging.getLogger(__name__)


class BaseMiddleware(object):
    """ Middleware base class for frameworks middleware hooks
    """

    def __init__(self, strategy, observation_queue, queue):
        LOGGER.debug("%s for %s", self.__class__.__name__, strategy)
        self.strategy = strategy
        self.observation_queue = observation_queue
        self.queue = queue

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.strategy))

    def execute_pre_callbacks(self, args=None, record_attack=True):
        """ Execute pre callbacks. Only process valid action, in this context it's raising.
        """
        action = execute_pre_callbacks(
            self.strategy.strategy_key, self.strategy, self, args=args
        )

        action_status = action.get("status")

        if not action_status:
            return args

        elif action_status == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            if record_attack:
                raise AttackBlocked(action.get("rule_name"))

        elif action_status == ACTIONS["ACTION_BLOCK"]:
            if record_attack:
                raise ActionBlock(action.get("action_id"))

        elif action_status == ACTIONS["ACTION_REDIRECT"]:
            if record_attack:
                raise ActionRedirect(action.get("action_id"), action["target_url"])

        elif action_status == ACTIONS["MODIFY_ARGS"]:
            return action.get("args")

        return args

    def execute_post_callbacks(self, response, args=None, record_attack=True):
        """ Execute post callbacks. Only process valid action, in this context it's raising.
        """
        action = execute_post_callbacks(
            self.strategy.strategy_key,
            self.strategy,
            self,
            result=response,
            args=args,
        )

        action_status = action.get("status")

        if not action_status:
            return response

        elif action_status == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            if record_attack:
                raise AttackBlocked(action.get("rule_name"))

        elif action_status == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")

        return response

    def execute_failing_callbacks(self, exception, args=None):
        """ Execute failing callbacks. Only process valid action, in this context it's None.
        """
        if isinstance(exception, Exception):
            exception = (exception.__class__, exception, None)

        action = execute_failing_callbacks(
            self.strategy.strategy_key,
            self.strategy,
            self,
            exc_info=exception,
            args=args,
        )

        action_status = action.get("status")
        if action_status and action_status == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Main runner module
"""

import logging
import sys
from time import time

from .deliverer import get_deliverer
from .ip_radix import Radix
from .list_filters import RequestPrefixListFilter
from .metrics import UnregisteredMetric
from .performance_metrics import PerformanceMetricsSettings
from .samplers import TracingSampler
from .signal_client import get_signal_client
from .utils import now

if sys.version_info[0] < 3:
    from Queue import Empty, Queue, Full
else:
    from queue import Empty, Queue, Full

MAX_QUEUE_LENGTH = 100
MAX_OBS_QUEUE_LENGTH = 1000

LOGGER = logging.getLogger(__name__)


class RunnerStop(object):
    """ Placeholder event for asking the runner to stop
    """


class MetricsEvent(object):
    """ Placeholder for asking observations aggregation to run
    """


class CappedQueue(object):
    """ Capped queue with opiniatied methods
    """

    def __init__(self, maxsize=None):
        if maxsize is None:
            maxsize = MAX_QUEUE_LENGTH

        self.maxsize = maxsize
        self.queue = Queue(self.maxsize)

    def get(self, timeout, block=True):
        """ Wait for up to timeout for an item to return and block while
        waiting
        """
        return self.queue.get(timeout=timeout, block=block)

    def get_nowait(self):
        """ Get without waiting, raise queue.Empty if nothing is present
        """
        return self.queue.get_nowait()

    def put(self, item):
        """ Tries to put an item to the queue, if the queue is empty, pop an
        item and try again
        """
        pushed = False
        while pushed is False:
            try:
                self.queue.put_nowait(item)
                pushed = True
            except Full:
                try:
                    self.queue.get_nowait()
                    self.queue.task_done()
                except Empty:
                    pass

    def clear(self):
        """ Clear all items in the queue
        """
        with self.queue.mutex:
            self.queue.queue.clear()

    def half_full(self):
        """ Return True if the current queue size if at least half the maxsize
        """
        return self.queue.qsize() > (self.maxsize / 2.)


class RunnerSettings(object):
    """ Various values that need to be shared across threads
    """

    def __init__(self):
        self.ips_whitelist = Radix(None)
        self.paths_whitelist = RequestPrefixListFilter()
        self.debug_flag = None
        self.tracing_identifier_prefix = None

    def set_ips_whitelist(self, addresses):
        """ Replace the current list of whitelisted IP networks
        """
        new_ips_whitelist = Radix(None)
        for address in addresses:
            new_ips_whitelist.insert(address, '/' in address)
        self.ips_whitelist = new_ips_whitelist

    def ips_whitelist_match(self, address):
        """ Return the first matching IP network, or None if no match could be
        found.
        """
        if address:
            return self.ips_whitelist.match(address)

    def set_paths_whitelist(self, paths):
        """ Replace the current list of whitelisted request path prefixes
        """
        self.paths_whitelist.reset(paths)

    def paths_whitelist_match(self, path):
        """ Return the first matching request path prefix, or None if no match
        could be found.
        """
        if path is not None:
            return self.paths_whitelist.match(path)

    def whitelist_match(self, request):
        """ Return the first matching IP network or request path prefix, or
        None if no match could be found.
        """
        ip = self.ips_whitelist_match(request.raw_client_ip)
        if ip is not None:
            return ip
        prefix = self.paths_whitelist_match(request.path)
        if prefix is not None:
            return prefix

    def set_debug_flag(self, flag):
        """ Set to True if the framework is in debug mode.
        """
        self.debug_flag = flag

    def get_debug_flag(self):
        """ Get the current debug flag.
        """
        return self.debug_flag

    def set_tracing_identifier_prefix(self, tracing_identifier_prefix):
        """ Set the current tracing idenftifier prefix.
        """
        self.tracing_identifier_prefix = tracing_identifier_prefix


def process_initial_commands(initial_payload, runner):
    """ Process the initial non-standard login payload
    """
    commands = initial_payload.get("commands", [])
    rulespack_id = initial_payload.get("pack_id", None)
    rules = initial_payload.get("rules", None)

    # Preprocess commands first
    for command in commands:
        if command["name"] == "instrumentation_enable" and not command["params"]:
            command["params"] = (rulespack_id, rules)

    # Process commands normally
    if commands:
        runner.process_commands(commands)

    actions = initial_payload.get("actions")
    runner.action_store.reload_from_dicts(actions or [])


class BaseRunner(object):
    """ Base runner class
    """


class Runner(BaseRunner):
    """ Main runner class

    Its job is to be the orchestrator and receiver for application communication
    It interacts with the backend through session, call heartbeat himself,
    execute commands and forward events
    """

    # Heartbeat delay is 5 minutes by default
    HEARTBEAT_DELAY = 300

    def __init__(
        self,
        queue,
        observation_queue,
        session,
        deliverer,
        remote_command,
        runtime_infos,
        instrumentation,
        action_store,
        metrics_store,
        settings,
        initial_features=None,
        performance_metrics=True,
        signal_client=None,
        tracing_sampler=None,
    ):
        self.logger = logging.getLogger(
            "{}.{}".format(self.__module__, self.__class__.__name__)
        )
        self.queue = queue
        self.observation_queue = observation_queue
        self.deliverer = deliverer
        self.remote_command = remote_command
        self.runtime_infos = runtime_infos
        self.instrumentation = instrumentation
        self.action_store = action_store
        self.metrics_store = metrics_store
        self.settings = settings
        self.stop = False
        # Used to globally disable performance metrics in the agent from the daemon
        self.performance_metrics = performance_metrics
        self.signal_client = signal_client
        self.tracing_sampler = tracing_sampler

        # Save the time runner started for checking warmup period termination
        self.started = time()

        if initial_features is None:
            initial_features = {}

        # The first time we shouldn't wait too long before sending heartbeat
        self.heartbeat_delay = initial_features.get(
            "heartbeat_delay", self.HEARTBEAT_DELAY
        )

        self.set_performance_metrics_settings(
            PerformanceMetricsSettings.from_features(initial_features)
        )

        self.budget = None

        self.whitelisted_metric = initial_features.get(
            "whitelisted_metric", True
        )

        self.last_heartbeat_request = 0

        self.session = session

        # Features
        self.logger.debug("Initial features %s", self.features_get())

        # Next things to send on next heartbeat
        self.next_commands_results = {}

    def run(self):
        """ Infinite loop
        """
        self.logger.debug("Starting the runner (%s)",
                          "with instrumentation enabled" if self.instrumentation.enabled
                          else "waiting for instrumentation")
        while self.stop is False:
            self.run_once()
        self.logger.debug("Exiting the runner now")

    def run_once(self, block=True):
        """ Tries to pop a message or send an heartbeat
        """
        try:
            event = self.queue.get(timeout=self.sleep_delay, block=block)
            self.logger.debug(
                "Run once, event: %s after %ss sleep delay",
                event,
                self.sleep_delay,
            )
            self.handle_message(event)

            # Exit now if should stop
            if self.stop:
                return
        except Empty:
            self.logger.debug(
                "No message after %ss sleep delay", self.sleep_delay
            )

        # Aggregate observations in transit in observations queue
        self.aggregate_observations()

        if self._should_do_heartbeat():
            self.do_heartbeat()

        # Tick the deliverer to publish batch if necessary
        self.deliverer.tick()

    @property
    def sleep_delay(self):
        """ Compute the sleeping delay by taking the heartbeat
        """
        return self.heartbeat_delay

    def handle_message(self, event):
        """ Handle incoming message
        Process RunnerStop message or pass event to the deliverer
        """
        if event is RunnerStop:
            self.logger.debug("RunnerStop found, logout")
            self.logout()
        elif event is MetricsEvent:
            self.aggregate_observations()
        else:
            self.deliverer.post_event(event)

    def process_commands(self, commands):
        """ handle commands
        """
        result = self.remote_command.process_list(commands, self)
        self.next_commands_results.update(result)

    def do_heartbeat(self):
        """ Do an heartbeat, publish finished metrics from MetricsStore and
        past commands results
        """
        payload = {
            "command_results": self.next_commands_results,
        }
        if self.signal_client is None:
            metrics = self.metrics_store.get_data_to_publish(
                now(), force_finalize=False)
            payload["metrics"] = [
                {k: v for k, v in metric.items() if k in ("name", "observation", "start", "finish")}
                for metric in metrics
            ]

        res = self.session.heartbeat(payload)
        self.last_heartbeat_request = time()

        # Clean sent command results
        self.next_commands_results = {}

        if res is not None:
            self.process_commands(res["commands"])

        if self.signal_client is not None:
            self.publish_metrics(force_finalize=False)

    def aggregate_observations(self):
        """ Empty the observation queue and update the metric store
        with the observations in the queue.
        """
        unregistered_metrics = set()
        try:
            while True:
                try:
                    name, at, key, value = self.observation_queue.get_nowait()
                    self.metrics_store.update(name, at, key, value)
                except UnregisteredMetric:
                    unregistered_metrics.add(name)
        except Empty:
            pass
        if unregistered_metrics:
            # metrics could be have been queued before
            # they were disabled by a feature change
            LOGGER.debug("Observations for unregistered metrics: %s",
                         ", ".join(unregistered_metrics))

    def publish_metrics(self, force_finalize=True):
        """ Publish finished metrics from MetricsStore
        """
        self.deliverer.post_metrics(
            self.metrics_store.get_data_to_publish(
                now(), force_finalize=force_finalize))

    def _should_do_heartbeat(self):
        """ Check if we should send an heartbeat because the delay is overdue
        """
        return (self.last_heartbeat_request + self.heartbeat_delay) < time()

    def logout(self):
        """ Run cleanup
        """
        self.logger.debug("Logout")

        # Flush metrics
        self.aggregate_observations()
        self.publish_metrics()

        # Drain deliverer
        self.deliverer.drain(resiliently=False)

        self.session.logout()

        # Mark itself as stopped if it's not the end of the application
        self.stop = True

    ###
    # Features
    ###

    def features_get(self):
        """ Returns the current values for all features switches
        """
        res = {
            "heartbeat_delay": self.heartbeat_delay,
            "batch_size": self.deliverer.batch_size,
            "use_signals": self.signal_client is not None,
            "max_staleness": self.deliverer.original_max_staleness,
            "call_counts_metrics_period": self.metrics_store.get_metric_period(
                "sqreen_call_counts"
            ),
            "whitelisted_metric": self.whitelisted_metric,
        }
        res.update(self.performance_metrics_settings.as_features())
        return res

    def set_heartbeat_delay(self, heartbeat_delay):
        """ Update the heartbeat delay
        """
        self.heartbeat_delay = heartbeat_delay

    def set_performance_cap_budget(self, budget):
        """ Update the budget of the performance cap feature.
        """
        if budget is not None and budget <= 0:
            raise ValueError("Expected a positive floating number %s" % budget)

        if budget is not None and self.performance_metrics_settings.enabled():
            LOGGER.info("Updated performance cap budget to %0.2fms", budget)
            self.metrics_store.register_metric(
                "request_overbudget_cb", "Sum",
                self.performance_metrics_settings.overtime_period
            )
            self.budget = budget
        else:
            LOGGER.info("Disabled performance cap")
            self.metrics_store.unregister_metric("request_overbudget_cb")
            self.budget = None

    def set_performance_metrics_settings(self, new_perf_met_set):
        self.performance_metrics_settings = new_perf_met_set
        if self.performance_metrics is not True:
            return
        if self.performance_metrics_settings.enabled():
            LOGGER.info("Updated performance monitoring settings")
            self.metrics_store.register_metric(
                "req", "Binning", self.performance_metrics_settings.period,
                base=self.performance_metrics_settings.base,
                unit=self.performance_metrics_settings.unit
            )
            self.metrics_store.register_metric(
                "sq", "Binning", self.performance_metrics_settings.period,
                base=self.performance_metrics_settings.base,
                unit=self.performance_metrics_settings.unit
            )
            self.metrics_store.register_metric(
                "pct", "Binning", self.performance_metrics_settings.period,
                base=self.performance_metrics_settings.pct_base,
                unit=self.performance_metrics_settings.pct_unit
            )
            for metric_name in self.metrics_store.list_metrics():
                # Reload performance monitoring for rules
                if metric_name.startswith("sq."):
                    self.metrics_store.register_metric(
                        metric_name, "Binning",
                        self.performance_metrics_settings.period,
                        base=self.performance_metrics_settings.base,
                        unit=self.performance_metrics_settings.unit
                    )
        else:
            LOGGER.info("Disabled performance monitoring")
            self.metrics_store.unregister_metric("req")
            self.metrics_store.unregister_metric("sq")
            self.metrics_store.unregister_metric("pct")
            for metric_name in self.metrics_store.list_metrics():
                if metric_name.startswith("sq."):
                    self.metrics_store.unregister_metric(metric_name)

    def set_call_counts_metrics_period(self, call_counts_metrics_period):
        self.metrics_store.register_metric(
            "sqreen_call_counts", "Sum", call_counts_metrics_period
        )

    def set_whitelisted_metric(self, enabled):
        self.whitelisted_metric = enabled

    def set_signal_client(self, batch_size, max_staleness, use_signals=True):
        if self.signal_client is not None:
            self.signal_client.flush()
            self.signal_client.close()

        self.signal_client = get_signal_client(
            self.session,
            batch_size=batch_size,
            max_staleness=max_staleness,
        ) if use_signals else None

    def set_deliverer(self, batch_size, max_staleness, use_signals=True):
        # Drain current deliverer
        self.deliverer.drain(resiliently=True)

        # Replace current deliverer
        signal_client = self.signal_client if use_signals else None
        self.deliverer = get_deliverer(batch_size, max_staleness, self.session,
                                       signal_client=signal_client)

    def set_ips_whitelist(self, addresses):
        """ Update current RunnerSettings ips_whitelist
        """
        self.logger.debug("Set IPs whitelist on settings %r", self.settings)
        self.settings.set_ips_whitelist(addresses)

    def set_paths_whitelist(self, paths):
        """ Update current RunnerSettings paths_whitelist
        """
        self.logger.debug("Set paths whitelist on settings %r", self.settings)
        self.settings.set_paths_whitelist(paths)

    def set_tracing_configuration(self, tracing_identifier_prefix, scope_definition):
        """ Update the tracing configuration
        """
        self.logger.debug("Set the tracing identifier prefix to %r", tracing_identifier_prefix)
        self.settings.set_tracing_identifier_prefix(tracing_identifier_prefix)
        if self.tracing_sampler is None:
            self.logger.debug("Creating a new tracing sampler with scope definition %r", scope_definition)
            self.tracing_sampler = TracingSampler(scope_definition)
        else:
            self.logger.debug("Updating tracing sampler scope definition %r", scope_definition)
            self.tracing_sampler.update_scope_definition(scope_definition)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Record request context."""

from logging import getLogger

from ..frameworks.aiohttp_framework import AioHTTPRequest, AioHTTPResponse
from .record_request_context import RecordRequestContext

LOGGER = getLogger(__name__)


class RecordRequestContextAioHTTP(RecordRequestContext):
    """Record request context."""

    def pre(self, instance, args, kwargs, **options):
        self._store_request(AioHTTPRequest(args[0]))

    def post(self, instance, args, kwargs, result=None, **options):
        self.storage.store_response(AioHTTPResponse(result))
        self._clear_request()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Matcher module."""

import re
from collections import defaultdict

from .utils import is_unicode


class Matcher(object):
    """Fast regex-like matcher."""

    __slots__ = ("patterns", "_regexes", "min_size")

    _STRING_OPTIONS = {
        "anywhere": ("", ""),
        "starts_with": ("^", ""),
        "ends_with": ("", "$"),
        "equals": ("^", "$"),
    }

    def __init__(self, patterns):
        self.patterns = patterns
        self._regexes = []
        self.min_size = None

        self._load_patterns(patterns)

    def _load_pattern(self, pattern):
        """Load a generic pattern object.

        Return a value (pattern, flags, size) where
        * pattern is the corresponding regex pattern
        * flags represent the bitwise flags of the regex
        * size is the minimum size of matching strings if it can be determined,
          None otherwise.
        """
        pattern_type = pattern["type"]
        value = pattern["value"]
        ignore_case = not pattern.get("case_sensitive", False)
        options = pattern.get("options", [])
        if pattern_type == "string":
            return self._load_string_pattern(
                pattern, value, ignore_case, options
            )
        elif pattern_type == "regexp":
            return self._load_regex_pattern(
                pattern, value, ignore_case, options
            )
        else:
            raise ValueError("unknown pattern type {!r}".format(pattern_type))

    def _load_string_pattern(self, pattern, value, ignore_case, options):
        """Load a string pattern object."""
        option = options[0] if options else "anywhere"
        if option not in self._STRING_OPTIONS:
            raise ValueError("unknown match function {!r}".format(option))
        re_start, re_end = self._STRING_OPTIONS[option]
        re_pattern = u"(?:{}{}{})".format(re_start, re.escape(value), re_end)
        re_flags = re.IGNORECASE if ignore_case else 0
        return re_pattern, re_flags, pattern.get("min_length", len(value))

    def _load_regex_pattern(self, pattern, value, ignore_case, options):
        """Load a regex pattern object."""
        re_flags = 0
        if ignore_case:
            re_flags |= re.IGNORECASE
        if "multiline" in options:
            re_flags |= re.MULTILINE
        return value, re_flags, pattern.get("min_length")

    def _load_patterns(self, patterns):
        """Load all patterns."""
        sized = True
        min_size = None
        re_pattern_map = defaultdict(list)
        for pattern in patterns:
            re_pattern, re_flags, size = self._load_pattern(pattern)
            re_pattern_map[re_flags].append(re_pattern)
            if not sized:
                continue
            if size is None:
                sized = False
            elif min_size is None or size < min_size:
                min_size = size
        self.min_size = min_size if sized else None
        self._regexes.extend(
            re.compile("|".join(re_patterns), re_flags)
            for re_flags, re_patterns in re_pattern_map.items()
        )

    def match(self, value):
        """Test if the string is matched by at least one of the patterns.

        This method accepts Python 2 str and unicode objects, and Python 3
        bytes and str objects.
        """
        # Correct encoding if possible and needed
        if not is_unicode(value):
            value = value.decode("utf-8", errors="replace")

        # Check length.
        if self.min_size is not None and len(value) < self.min_size:
            return False

        # Match regexes.
        for regex in self._regexes:
            if regex.search(value) is not None:
                return True

        return False



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import sys

from .__about__ import __version__
from .accumulator import BatchingAccumulator
from .compat_model import Signal, SignalType, Trace
from .sender import SyncSender

if sys.version_info[0] >= 3:
    from concurrent.futures import ThreadPoolExecutor
else:
    from sqreen._vendors.concurrent.futures import ThreadPoolExecutor

if sys.version_info >= (3, 5):
    from typing import Any, Optional

    from .compat_model import AnySignal


class SyncClient(object):
    """Send signals to the Sqreen Ingestion service.

    :param token: Your application API token.
    :param app_name: (optional) Your application name.
    :param proxy_url: (optional) Send requests througth this proxy.
    :param max_batch_size: (optional) Maximum number of items sent per batch (default to 50).
    :param interval_batch: (optional) Interval at which non-empty batch should be sent (default to 60s).
    :param session_token: (optional) When true, token is a session token instead of an API token.
    :param base_url: (optional) Set a different ingestion API URL.
    """

    accumulator_class = BatchingAccumulator
    sender_class = SyncSender

    user_agent = "sqreen-python-security-signal-sdk/{}".format(__version__)
    max_workers = 2

    def __init__(self, token, app_name=None, proxy_url=None, max_batch_size=50,
                 interval_batch=60, session_token=False, base_url=None):
        # type: (str, Optional[str], Optional[str], int, float, bool, Optional[str]) -> None

        headers = {"User-Agent": self.user_agent}
        if session_token:
            headers["X-Session-Key"] = token
        else:
            headers["X-Api-Token"] = token
            if app_name is not None:
                headers["X-App-Name"] = app_name

        self.sender = self.sender_class(base_url=base_url, proxy_url=proxy_url, headers=headers)
        self.accumulator = self.accumulator_class(
            max_batch_size=max_batch_size, linger_time=interval_batch)
        self.executor = ThreadPoolExecutor(max_workers=self.max_workers)

    def point(self, signal_name, payload, **properties):  # type: (str, Any, **Any) -> None
        """Record a point signal to be sent."""
        properties["type"] = SignalType.POINT
        return self.signal(signal_name, payload, **properties)

    def metric(self, signal_name, payload, **properties):  # type: (str, Any, **Any) -> None
        """Record a metric signal to be sent."""
        properties["type"] = SignalType.METRIC
        return self.signal(signal_name, payload, **properties)

    def signal(self, signal_name, payload, **properties):  # type: (str, Any, **Any) -> None
        """Record a signal to be sent."""
        signal = dict(signal_name=signal_name, payload=payload)  # type: Signal
        signal.update(properties)  # type: ignore
        return self._add_and_send(signal)

    def trace(self, data, **properties):  # type: (Any, **Any) -> None
        """Record a trace to be sent."""
        trace = dict(data=data)  # type: Trace
        trace.update(properties)  # type: ignore
        return self._add_and_send(trace)

    def _add_and_send(self, data):  # type: (AnySignal) -> None
        batch = self.accumulator.add(data)
        if batch:
            self.executor.submit(self.sender.send_batch, batch)

    def flush(self, soft=False):  # type: (bool) -> None
        """Send all pending signals and traces.

        :param soft: (optional) Do not send the batch if it has not exceeded
        the interval time.
        """
        batch = self.accumulator.flush(soft=soft)
        if batch:
            self.executor.submit(self.sender.send_batch, batch)

    def close(self):  # type: () -> None
        """Close the client.
        """
        self.executor.shutdown(wait=True)
        self.sender.close()


Client = SyncClient



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Callback for tracing outgoing urllib requests
"""
import uuid

from ..constants import ACTIONS
from .trace import Trace


class TraceOutgoingUrllib(Trace):

    # Hook http.client.HTTPConnection._send_request

    def pre(self, instance, args, kwargs, **options):
        try:
            host_ip, host_port = instance.sock.getpeername()
        except Exception:
            host_ip, host_port = None, None

        tracing_identifier_prefix = self.runner.settings.tracing_identifier_prefix
        if tracing_identifier_prefix is not None:
            request_identifier = uuid.uuid4()
            tracing_identifier = "{}.{}".format(tracing_identifier_prefix, request_identifier)
        else:
            tracing_identifier = None
            request_identifier = None

        options["host_ip"] = host_ip
        options["host_port"] = host_port
        options["tracing_identifier_prefix"] = tracing_identifier_prefix
        options["tracing_identifier"] = tracing_identifier
        options["request_identifier"] = request_identifier

        super(TraceOutgoingUrllib, self).pre(instance, args, kwargs, **options)

        options = self.data.get("options", {})
        if options.get("propagate_tracing_identifier", False) and tracing_identifier is not None:
            method, url, body, headers, encode_chunked = args
            new_headers = dict(headers)
            header_name = options.get("header_name", "X-Sqreen-Trace-Identifier")
            new_headers[header_name] = tracing_identifier
            return {
                "status": ACTIONS["MODIFY_ARGS"],
                "args": ((method, url, body, new_headers, encode_chunked), {}),
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" uWSGI webserver module
"""
from .base import BaseWebServer

WHITELISTED_OPTS = [
    "enable-threads",
    "lazy-apps",
    "master",
    "py-call-osafterfork",
]


class UWSGIWebServer(BaseWebServer):
    """uWSGI webserver."""

    def __init__(self, module):
        self._module = module

    def various_infos(self):
        """Gather various information about uwsgi."""
        opts = getattr(self._module, "opt", {})
        return {
            "name": "uwsgi",
            "version": getattr(self._module, "version", "unknown"),
            "options": {opt: opts.get(opt) for opt in WHITELISTED_OPTS
                        if opt in opts},
        }

    @classmethod
    def detect(cls):
        """Detect if uwsgi module is present."""
        try:
            import uwsgi  # type: ignore
        except ImportError:
            return None
        return cls(module=uwsgi)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Authentication SDK functions
"""
from ..utils import HAS_TYPING

if HAS_TYPING:
    from typing import Mapping, Optional, Union

    UserIdentifier = Union[str, int]
    UserDict = Mapping[str, UserIdentifier]


def auth_track(success, **user_identifiers):
    # type: (bool, **UserIdentifier) -> None
    """ Register a successfull or failed authentication attempt (based on
    success boolean) for an user identified by the keyword-arguments.
    For example:

    Register a successfull authentication attempt for user with email
    "foobar@example.com":

    >>> auth_track(True, email="foobar@example.com")

    Register a failed authentication attempt for user with id 42:

    >>> auth_track(False, user_id=42)
    """


def signup_track(**user_identifiers):
    # type: (**UserIdentifier) -> None
    """ Register a new account signup identified by the keyword-arguments.
    For example:

    Register a new account signup for user identifed by email "foobar@example.com" or
    bu user id 42:

    >>> auth_track(email="foobar@example.com", user_id=42)
    """


def identify(user_identifiers, traits=None):
    # type: (UserDict, Optional[Mapping]) -> None
    """ Associate the current request with the account identified by
    the identifier.

    Example:

    >>> identify({"user_id": 42, "email": "foobar@example.com"})
    """



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Insert custom headers for Flask
"""
import logging

from .headers_insert import BaseHeadersInsertCB

LOGGER = logging.getLogger(__name__)


class HeadersInsertCBFlask(BaseHeadersInsertCB):
    """ Callback that add the custom sqreen header
    """

    def post(self, instance, args, kwargs, result=None, **options):
        """ Set headers
        """
        try:
            response = result
            for header_name, header_value in self.headers.items():
                response.headers.set(header_name, header_value)
        except Exception:
            LOGGER.warning("An error occurred", exc_info=True)

        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Asynchronous variant of import_hook."""

from ..async_hook_point import async_hook_point
from .import_hook import ImportHookStrategy


class AsyncImportHookStrategy(ImportHookStrategy):
    """Asynchronous variant of ImportHookStrategy."""

    def import_hook_callback(self, original):
        return async_hook_point(self, self.hook_path, self.hook_name, original)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Custom error page for aiohttp."""

from logging import getLogger

from .headers_insert import convert_to_str
from .sqreen_error_page import BaseSqreenErrorPage

LOGGER = getLogger(__name__)


class SqreenErrorPageAioHTTP(BaseSqreenErrorPage):
    """Custom error page for aiohttp."""

    def pre(self, instance, args, kwargs, **options):

        exc = kwargs.get("exc") or args[2]
        ret = self.handle_exception(exc)

        if ret is not None:
            status_code, content, headers = ret

            from aiohttp.web_response import Response  # type: ignore

            content_type = headers.get("Content-Type", "text/html")
            resp = Response(status=status_code, content_type=content_type,
                            body=content.encode("utf-8"))
            for header_name, header_value in convert_to_str(headers.items()):
                resp.headers[header_name] = header_value
            return {
                "status": "override",
                "new_return_value": resp
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Enforce user actions with SDK callbacks
"""
import logging

from ..rules import RuleCallback
from ..utils import now

LOGGER = logging.getLogger(__name__)


class SDKIdentify(RuleCallback):

    def pre(self, instance, args, kwargs, **options):
        """Associate the current request with a user."""
        traits = kwargs.get("traits") or {}
        user_identifiers = {}
        if args:
            user_identifiers.update(args[0])

        self.storage.observe(
            "sdk", ["identify", now(), user_identifiers, traits],
            report=False
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Record request context
"""
import sys
from io import BytesIO
from logging import getLogger

from ..constants import ACTIONS
from ..rules import RuleCallback

LOGGER = getLogger(__name__)


class WSGIInputWrapper:
    """
    Wrap wsgi.input and keep a preview of the read bytes.
    """

    DEFAULT_MAX_PREVIEW_LENGTH = 4096

    def __init__(self, rule, original, length_hint=None,
                 max_preview_length=None):
        self.rule = rule
        self.original = original
        self.original_read = 0
        self.original_length_hint = length_hint \
            if length_hint is not None else sys.maxsize
        self.preview_max_length = max_preview_length \
            if max_preview_length is not None \
            else self.DEFAULT_MAX_PREVIEW_LENGTH
        self.preview_remaining = self.preview_max_length
        self.buf = BytesIO()

    def preview_callback(self, preview, end_of_input):
        """
        Callback to get the WSGI input preview when ready.
        """
        LOGGER.debug("Got a preview of %d bytes (end of input %r)",
                     len(preview), end_of_input)
        self.rule.storage.store_request_input_preview(preview, end_of_input)

    def _feed_preview(self, data, size_hint=0):
        if not self.preview_remaining:
            return
        try:
            self.original_read += len(data)
            preview = data[:self.preview_remaining]
            self.buf.write(preview)
            self.preview_remaining -= len(preview)
            # End of stream or preview is ready
            end_of_input = not data or size_hint is None \
                or size_hint < 0 or len(data) < size_hint \
                or self.original_read >= self.original_length_hint
            if not self.preview_remaining or end_of_input:
                try:
                    self.preview_callback(self.buf.getvalue(), end_of_input)
                finally:
                    self.preview_remaining = 0
                    # Don't keep the reference will free the buf memory
                    self.buf = None
        except Exception as e:
            self.rule.record_exception(e, sys.exc_info())

    def read(self, *args, **kwargs):
        data = self.original.read(*args, **kwargs)
        size = args[0] if args else kwargs.get("size")
        self._feed_preview(data, size_hint=size)
        return data

    def readline(self, *args, **kwargs):
        data = self.original.readline(*args, **kwargs)
        self._feed_preview(data)
        return data

    def readlines(self, *args, **kwargs):
        it = self.original.readlines(*args, **kwargs)
        for data in it:
            self._feed_preview(data)
            yield data
        else:
            self._feed_preview(b"")

    def __iter__(self):
        for data in self.original:
            self._feed_preview(data)
            yield data
        else:
            self._feed_preview(b"")


class WSGIInputPreview(RuleCallback):
    """ Get a preview of the WSGI input stream.
    """

    INTERRUPTIBLE = False

    def pre(self, instance, args, kwargs, **options):
        """ Replace the WSGI input stream with our wrapper to get a
        preview of the body.
        """
        environ = args[0]
        input = environ.get("wsgi.input")
        current_request = self.storage.get_current_request()
        if input is not None and current_request is not None \
                and not isinstance(input, WSGIInputWrapper):
            max_preview_length = self.data.get("max_preview_length")
            input_wrapper = WSGIInputWrapper(
                self, input, current_request.content_length,
                max_preview_length)
            environ = dict(environ.items())
            environ["wsgi.input"] = input_wrapper
        return {
            "status": ACTIONS["MODIFY_ARGS"],
            "args": ([environ], kwargs)
        }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import json
import logging
import sys

from sqreen._vendors.urllib3 import Retry, poolmanager, util  # type: ignore
from sqreen._vendors.urllib3.exceptions import HTTPError, MaxRetryError  # type: ignore
from sqreen._vendors.urllib3.util import Timeout

from .compat_model import Batch, Signal, Trace
from .exceptions import (AuthenticationFailed, DataIngestionFailed,
                         UnexpectedStatusCode)
from .utils import CustomJSONEncoder, reencode_payload

if sys.version_info[0] >= 3:
    from urllib import parse as urlparse
else:
    import urlparse

if sys.version_info >= (3, 5):
    from typing import Any, Mapping, Optional, Union, Type

    from .compat_model import AnySignal


LOGGER = logging.getLogger(__name__)


class BaseSender(object):
    """Base sender for the Sqreen Ingestion service.

    :param base_url: (optional) URL of the Ingestion service.
    :param proxy_url: (optional) URL of a Proxy server.
    :param headers: (optional) Headers to send with all requests.
    :param json_encoder: (optional) JSON serializer for data to be sent.
    """

    default_base_url = "https://ingestion.sqreen.com/"  # type: str
    default_json_encoder = CustomJSONEncoder

    def __init__(self, base_url=None, proxy_url=None, headers={}, json_encoder=None):
        # type: (Optional[str], Optional[str], Mapping[str, str], Optional[Type[json.JSONEncoder]]) -> None
        self.base_url = base_url or self.default_base_url
        self.proxy_url = proxy_url
        self.headers = headers
        self.json_encoder = json_encoder or self.default_json_encoder

    def send_batch(self, data, headers={}, **kwargs):
        # type: (Batch, Mapping[str, str], **Any) -> None
        return self.send("/batches", data, headers=headers, **kwargs)

    def send_signal(self, data, headers={}, **kwargs):
        # type: (Signal, Mapping[str, str], **Any) -> None
        return self.send("/signals", data, headers=headers, **kwargs)

    def send_trace(self, data, headers={}, **kwargs):
        # type: (Trace, Mapping[str, str], **Any) -> None
        return self.send("/traces", data, headers=headers, **kwargs)

    def send(self, endpoint, data, headers={}, **kwargs):
        # type: (str, Union[AnySignal, Batch], Mapping[str, str], **Any) -> None
        raise NotImplementedError

    def serialize_data(self, data):
        # type: (Union[AnySignal, Batch]) -> str
        try:
            return json.dumps(
                data, separators=(",", ":"), cls=self.json_encoder)
        except UnicodeDecodeError:
            reencoded_data = reencode_payload(data)
            return json.dumps(
                reencoded_data, separators=(",", ":"), cls=self.json_encoder)

    def handle_response(self, response):
        if response.status not in (200, 202):
            if response.status == 422:
                raise DataIngestionFailed
            elif response.status in (401, 403):
                raise AuthenticationFailed
            raise UnexpectedStatusCode(response.status)
        # ignore the response content for now

    def close(self):  # type: () -> None
        raise NotImplementedError


class SyncSender(BaseSender):
    """
    Sender based on urllib3 for the Ingestion service.

    :param base_url: (optional) URL of the Ingestion service.
    :param proxy_url: (optional) URL of a Proxy server.
    :param headers: (optional) Headers to send with all requests.
    :param json_encoder: (optional) JSON serializer for data to be sent.
    """

    max_pool_size = 10
    retry_policy = Retry(
        total=3,
        method_whitelist=False,
        status_forcelist={500, 502, 503, 504, 408},
        raise_on_status=True,
    )
    timeout_policy = Timeout(connect=10, read=10)

    def __init__(self, base_url=None, proxy_url=None, headers={}, json_encoder=None):
        # type: (Optional[str], Optional[str], Mapping[str, str], Optional[Type[json.JSONEncoder]]) -> None
        base_headers = util.make_headers(keep_alive=True, accept_encoding=True)
        base_headers.update(headers)
        super(SyncSender, self).__init__(
            base_url=base_url, proxy_url=proxy_url, headers=base_headers,
            json_encoder=json_encoder)

        options = dict(
            block=True,
            maxsize=self.max_pool_size,
        )
        if self.proxy_url is not None:
            self.pool_manager = poolmanager.ProxyManager(self.proxy_url, **options)  # type: poolmanager.PoolManager
        else:
            self.pool_manager = poolmanager.PoolManager(**options)

    def _url(self, endpoint):
        return urlparse.urljoin(self.base_url, endpoint, allow_fragments=False)

    def send(self, endpoint, data, headers={}, **kwargs):
        # type: (str, Union[AnySignal, Batch], Mapping[str, str], **Any) -> None
        assert self.pool_manager is not None
        body = self.serialize_data(data)
        request_headers = dict(self.headers)
        request_headers["Content-Type"] = "application/json"
        request_headers.update(headers)
        url = self._url(endpoint)
        try:
            response = self.pool_manager.urlopen(
                "POST",
                url,
                body=body,
                headers=request_headers,
                preload_content=True,
                release_conn=True,
                redirect=False,
                retries=self.retry_policy,
                timeout=self.timeout_policy,
                **kwargs
            )
        except (MaxRetryError, HTTPError) as exc:
            LOGGER.info(
                "Couldn't connect to %s due to exception %s", url, type(exc).__name__
            )
        else:
            return self.handle_response(response)

    def close(self):  # type: () -> None
        self.pool_manager.clear()
        del self.pool_manager


Sender = SyncSender



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Record a WSGI response.
"""

from logging import getLogger

from ..frameworks.wsgi import WSGIResponse
from ..rules import RuleCallback

LOGGER = getLogger(__name__)


class RecordResponse(RuleCallback):

    INTERRUPTIBLE = False

    def pre(self, instance, args, kwargs, **options):
        """ Record the WSGI response if empty.
        """
        current_request = self.storage.get_current_request()
        # Do not set a response if we don't have a request yet.
        if current_request is not None:
            response = WSGIResponse(*args[:2])
            self.storage.store_response_default(response)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import sys
import threading
import time

from .compat_model import Batch

if sys.version_info >= (3, 5):
    from typing import Optional

    from .compat_model import AnySignal


class BatchingAccumulator(object):
    """Accumulate signals into a batch.

    :param max_batch_size: (optional) Maximum number of items in the batch (default to 50).
    :param linger_time: (optional) Maximum age of a non-empty batch in seconds (default to 60s).
    """

    def __init__(self, max_batch_size=50, linger_time=60):
        # type: (int, float) -> None
        self.max_batch_size = max_batch_size
        self.linger_ms = int(linger_time * 1000)
        self.batch = Batch()
        self.batch_creation_time = 0
        self.batch_lock = threading.RLock()

    def add(self, signal):  # type: (AnySignal) -> Optional[Batch]
        """Add a signal to the current batch and flush it if needed."""
        with self.batch_lock:
            if not self.batch:
                self.batch_creation_time = self._current_time_ms()
            self.batch.append(signal)
            return self.flush(soft=True)

    def flush(self, soft=False):  # type: (bool) -> Optional[Batch]
        """Flush the current batch if not empty.

        :param soft: (optional) Do not flush the batch if it has not exceeded
        the linger time.
        """
        with self.batch_lock:
            if soft:
                now = self._current_time_ms()
                soft = len(self.batch) < self.max_batch_size \
                    and (now - self.batch_creation_time) < self.linger_ms
            if not soft and self.batch:
                batch = self.batch
                self.batch = Batch()
                return batch
            return None

    @staticmethod
    def _current_time_ms():  # type: () -> int
        return int(time.time() * 1000)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Import hook related helpers
"""
import logging
import sys
import threading

if sys.version_info >= (3, 4):
    from importlib.util import find_spec
    PY3_IMPORT_HOOK = True
else:
    find_spec = None
    PY3_IMPORT_HOOK = False

LOGGER = logging.getLogger(__name__)


def get_hook_parent(module, hook_class):
    """ From a module and a class, retrieve the right class
    and returns both the hook parent and the full hook path
    """
    if hook_class:
        if "." in hook_class:
            raise NotImplementedError()
        else:
            return getattr(module, hook_class, None)
    else:
        return module


def get_module_parent(module_name):
    """ Return the parent module name

    >>> get_module_parent('django.conf')
    'django'
    >>> get_module_parent('django')
    'django'
    """
    splitted = module_name.split(".")
    if len(splitted) == 1:
        return module_name
    else:
        return ".".join(splitted[:-1])


def get_original(module, module_class_name, module_object_name):
    """ Return the object in module at path module_class_name / module_object_name

    For example: 'itertools:None:count' would target the function named count
        in the itertools module.
        And 'django.core.handlers.wsgi:WSGIHandler:__call__' would target the
        method '__call__' of the class 'WSGIHandler' in the module
        'django.core.handlers.wsgi'.
    """
    # Get module class
    module_class = get_hook_parent(module, module_class_name)

    if module_class is None:
        LOGGER.info("Module %r has no attribute %r", module, module_class_name)
        return None

    # Get module object
    module_object = get_hook_parent(module_class, module_object_name)

    if module_object is None:
        LOGGER.info(
            "Module class %r::%r (%r) has no attribute %r",
            module,
            module_class_name,
            module_class,
            module_object_name,
        )
        return None

    return module_object


def update_original(module, patched, module_class_name, module_object_name):
    """ Update the object at the path module_class_name + module_object in the
    module with patched.
    The paths are the same that get_original.
    """
    module_class = get_hook_parent(module, module_class_name)
    setattr(module_class, module_object_name, patched)


class ExistingPatchOnSameObject(Exception):
    """ Raise in case a patcher is already present for the same object
    """

    pass


class ModuleFinder(object):
    """ Custom import hook for Python 2/3 that import a module and hook on several
    hook points of this module
    """

    def __init__(self):
        self.patchers = {}
        self.loading = {}
        # The ModuleFinder is used across all threads so we must lock the
        # patching operation to prevent race conditions. In fact, a thread can be
        # loading a module while a second tries to patch it. It leads to random errors
        # such as "module has no attribute".
        # The lock is recursive because the find_module method calls itself.
        self.lock = threading.RLock()

    def register_patcher(
        self, module_name, module_class, module_object, patcher
    ):
        """ Register a patcher for a specific path, a path is composed of
        module_name, module_class and module_object.
        For example: 'itertools:None:count' would target the function named count
        in the itertools module.
        And 'django.core.handlers.wsgi:WSGIHandler:__call__' would target the
        method '__call__' of the class 'WSGIHandler' in the module
        'django.core.handlers.wsgi'.

        There could be only one patcher per path, if a patcher is already
        registered for a path, a ExistingPatchOnSameObject exception will be
        raised.
        """
        path = "%s:%s:%s" % (module_name, module_class, module_object)
        LOGGER.debug("Adding patcher for path %s %r", path, patcher)
        patchers = self.patchers.setdefault(module_name, {}).setdefault(
            module_class, {}
        )

        if module_object in patchers:
            raise ExistingPatchOnSameObject("Patcher already here")

        patchers[module_object] = {"patcher": patcher, "patched": False}

    def inject(self):
        """ Inject the global import hook in sys.meta_path if it's not already
        present
        """
        if sys.meta_path[0] is not self:
            sys.meta_path.insert(0, self)

    def apply_patchers(self):
        """ Try to apply not already applied patchers on modules already present
        in sys.modules as they would not be applied otherwise.
        """
        # Block the patching operation if there are still modules loading.
        with self.lock:
            for module_name, _ in self.patchers.items():

                if module_name in sys.modules:
                    msg = "Module %s was already imported, hooking now"
                    LOGGER.warning(msg, module_name)

                    self.patch_module(module_name, sys.modules[module_name])

    def patch_module(self, fullname, module):
        """ For a module named fullname, apply all patchers defined for it
        """
        for module_class, module_objects in self.patchers.get(
            fullname, {}
        ).items():
            for module_object, patcher in module_objects.items():
                if patcher["patched"] is False:
                    self._patch(module, module_class, module_object, patcher)
                else:
                    path = "%s:%s:%s" % (fullname, module_class, module_object)
                    LOGGER.debug(
                        "Patcher %r already applied for %s",
                        patcher["patcher"],
                        path,
                    )

        return module

    def _patch(self, module, module_class, module_object, patcher):
        """ Apply a single patcher on a single path
        """
        # Retrieve the original method
        original = get_original(module, module_class, module_object)
        if original is not None:
            # Patch it
            patched = patcher["patcher"](original)
            # And update back the module
            update_original(module, patched, module_class, module_object)
        # Mark as already patched
        patcher["patched"] = True

    def find_module(self, fullname, path=None):
        """ Import hook main API, called when the application try to import
        every module not already in sys.modules.
        """
        # Early exit
        if fullname not in self.patchers:
            return

        # Lock because we are currently loading a module.
        with self.lock:
            is_loading = fullname in self.loading
            LOGGER.debug(
                "Finding module %s, loading? %s, %r",
                fullname,
                is_loading,
                sys.meta_path,
            )
            # Avoid recursion as we import the module in a way that recheck every
            # import hook
            if is_loading:
                return

            self.loading[fullname] = True
            try:

                # In python 2, call back the whole import mechanism but ignoring
                # ourself
                if PY3_IMPORT_HOOK is False:
                    __import__(fullname)

                    modules_keys = sys.modules
                    if (
                        fullname in modules_keys
                        or get_module_parent(fullname) in modules_keys
                    ):
                        return self
                # In python 3, call loader
                else:
                    spec = find_spec(fullname, path)
                    if spec is not None and spec.loader is not None:
                        return ModuleLoader(spec.loader, self)

                LOGGER.debug("Module %s was not loaded", fullname)

            finally:
                del self.loading[fullname]

    def load_module(self, fullname):
        """ Load the module either from sys.modules or from its parent module.
        Used only in Python 2.
        """
        # Block the patching while we are loading a module.
        with self.lock:
            if fullname in sys.modules:
                module = sys.modules[fullname]
            else:
                parent_module_name = get_module_parent(fullname)
                if parent_module_name in sys.modules:
                    parent_module = sys.modules[parent_module_name]

                    module = getattr(parent_module, fullname.split(".")[-1])

            return self.patch_module(fullname, module)


class ModuleLoader(object):
    """ A module loader for Python 3
    """

    def __init__(self, loader, finder):
        self.loader = loader
        self.finder = finder

    def load_module(self, fullname):
        """ Import hook in python 3 call for loading a module
        """
        # Block the ModuleFinder from patching while we are loading a module.
        with self.finder.lock:
            module = self.loader.load_module(fullname)
            return self.finder.patch_module(fullname, module)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Low-level HTTP interaction with sqreen API
"""
import codecs
import json
import os
import sys
from datetime import datetime
from json import JSONEncoder

from .__about__ import __version__
from ._vendors.urllib3 import (  # type: ignore
    PoolManager,
    ProxyManager,
    Retry,
    make_headers,
)
from .exceptions import (
    InvalidJsonResponse,
    InvalidResponseContentType,
    InvalidStatusCodeResponse,
    StatusFailedResponse,
)
from .utils import Iterable, Mapping, is_string, is_unicode

if sys.version_info[0] < 3:
    import urlparse
else:
    import urllib.parse as urlparse


IS_PYTHON_3 = sys.version_info.major == 3


class SqreenJSONEncoder(JSONEncoder):
    """ Custom JsonEncoder which can handle datetime objects
    >>> from datetime import datetime
    >>> date = datetime(2016, 3, 22, 14, 43, 14, 742306)
    >>> json.dumps(date, cls=SqreenJSONEncoder)
    '"2016-03-22T14:43:14.742306"'
    >>> json.dumps(2 + 1j, cls=SqreenJSONEncoder)
    '"(2+1j)"'
    >>> class Fail(object):
    ...     def __repr__(self):
    ...         raise NotImplementedError("Oups")
    ...
    >>> f = Fail()
    >>> json.dumps(f, cls=SqreenJSONEncoder)
    '"instance of type <class \\'sqreen.http_client.Fail\\'>"'
    """

    def default(self, obj):

        if isinstance(obj, datetime):
            return obj.isoformat()

        # Process bytes for python3
        if IS_PYTHON_3 and isinstance(obj, bytes):
            return obj.decode("utf-8", errors="ascii_to_hex")

        # For unknows types, JSONEncoder will raise a TypeError, instead,
        # returns obj repr
        try:
            return repr(obj)
        # If failing, return a fallback string here with the class name
        except Exception:
            return "instance of type {}".format(repr(obj.__class__))


def ascii_to_hex(exception):
    """ On unicode decode error (bytes -> unicode error), tries to replace
    invalid unknown bytes by their hex notation.
    """
    if isinstance(exception, UnicodeDecodeError):
        obj = exception.object
        start = exception.start
        end = exception.end

        invalid_part = obj[start:end]
        result = []

        for character in invalid_part:
            # Python 2 strings
            if isinstance(character, str):
                result.append(u"\\x{}".format(character.encode("hex")))
            # Python 3 int
            elif isinstance(character, int):
                result.append(u"\\{}".format(hex(character)[1:]))
            else:
                raise exception
        result = ("".join(result), end)
        return result

    raise exception


codecs.register_error("ascii_to_hex", ascii_to_hex)


def encode_payload(payload):
    """ Tries to encode a payload into JSON, if not working, reencode the
    payload and retry to encode it into JSON.
    """
    try:
        return json.dumps(
            payload, separators=(",", ":"), cls=SqreenJSONEncoder
        )
    except UnicodeDecodeError:
        reencoded_payload = reencode_payload(payload)
        return json.dumps(
            reencoded_payload, separators=(",", ":"), cls=SqreenJSONEncoder
        )


def reencode_payload(payload):
    """ Do everything necessary to be able to encode a payload into JSON
    """
    if is_string(payload):
        return _reencode_string(payload)
    elif isinstance(payload, Mapping):
        return {
            _reencode_string(key): reencode_payload(value)
            for key, value in payload.items()
        }
    elif isinstance(payload, Iterable):
        return [reencode_payload(item) for item in payload]
    else:
        return payload


def _reencode_string(string):
    """ Ensure that the string is encodable into JSON
    """
    if not is_unicode(string):
        return string.decode("utf-8", errors="ascii_to_hex")

    return string


def where_crt():
    """ Returns the path of the bundled crt
    """
    current_file = os.path.split(__file__)[0]
    return os.path.join(current_file, "ca.crt")


USER_AGENT = "sqreen-python/{}".format(__version__)


class Urllib3Connection(object):
    """ Class responsible for making http request to sqreen API,

    handle connection pooling, retry and inbound/outbound formatting
    """

    RETRY_STATUS = {500, 502, 503, 504, 408}
    # Beware that the retry policy will block the Sqreen thread until
    # the request is successful.
    RETRY = Retry(
        total=5,
        method_whitelist=False,
        status_forcelist=RETRY_STATUS,
        backoff_factor=1,
    )
    RETRY_LONG = Retry(
        total=15,
        method_whitelist=False,
        status_forcelist=RETRY_STATUS,
        backoff_factor=1,
    )
    RETRY_ONCE = Retry(
        total=1,
        method_whitelist=False,
        status_forcelist=RETRY_STATUS,
        backoff_factor=1,
    )

    PATH_PREFIX = "/sqreen/"

    def __init__(self, server_url, proxy_url=None, user_agent=USER_AGENT):
        self.server_url = server_url
        self.proxy_url = proxy_url
        self.parsed_server_url = urlparse.urlparse(server_url)
        self.user_agent = user_agent

        if self.proxy_url:
            self.connection = ProxyManager(
                proxy_url,
                cert_reqs="CERT_REQUIRED",
                ca_certs=where_crt(),
                timeout=60,
            )
        else:
            self.connection = PoolManager(
                cert_reqs="CERT_REQUIRED", ca_certs=where_crt(), timeout=60
            )

    def get_user_agent(self):
        return getattr(self, "_user_agent", USER_AGENT)

    def set_user_agent(self, value):
        self.base_headers = make_headers(
            keep_alive=True, accept_encoding=True, user_agent=value
        )
        self._user_agent = value

    user_agent = property(get_user_agent, set_user_agent)

    def _url(self, url):
        """ Format and return an url based on instance path prefix
        """
        path = urlparse.urljoin(self.PATH_PREFIX, url)
        return urlparse.urlunparse(
            (
                self.parsed_server_url[0],
                self.parsed_server_url[1],
                path,
                None,
                None,
                None,
            )
        )

    def post(self, endpoint, data=None, headers=None, retries=None):
        """ Post a request to the backend
        """
        url = self._url(endpoint)

        if headers is None:
            headers = {}

        if data is not None:
            data = encode_payload(data)
            headers["Content-Type"] = "application/json"

        # Add base headers
        headers.update(self.base_headers)

        response = self.connection.urlopen(
            "POST", url, headers=headers, body=data, retries=retries
        )

        return self._parse_response(response)

    def get(self, endpoint, headers=None, retries=None):
        """ Get an endpoint in the backend
        """
        url = self._url(endpoint)

        if headers is None:
            headers = {}

        # Add base headers
        headers.update(self.base_headers)

        response = self.connection.urlopen(
            "GET", url, headers=headers, retries=retries
        )

        return self._parse_response(response)

    def _parse_response(self, response):
        """ Try to decode response body to json
        """
        if response.status < 200 or response.status >= 300:
            raise InvalidStatusCodeResponse(response.status, response.data)

        if response.headers.get("Content-Type") != "application/json":
            raise InvalidResponseContentType(
                response.headers.get("Content-Type")
            )

        try:
            json_response = json.loads(response.data.decode("utf-8"))
        except ValueError as exc:
            raise InvalidJsonResponse(exc)

        if json_response.get("status", False) is False:
            raise StatusFailedResponse(json_response)

        return json_response



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import os.path
from distutils.command.build_ext import build_ext
from distutils.errors import (
    CCompilerError,
    DistutilsExecError,
    DistutilsPlatformError,
)

from setuptools import Extension, find_packages, setup

root_dir = os.path.abspath(os.path.dirname(__file__))
metadata = {}

about_file = os.path.join(root_dir, "sqreen", "__about__.py")
with open(about_file, "rb") as f:
    exec(f.read(), metadata)

readme_file = os.path.join(root_dir, "README.md")
with open(readme_file, "rb") as f:
    long_description = f.read().decode("utf-8")

base_setup = dict(
    name="sqreen",
    version=metadata["__version__"],
    description="Sqreen agent to protect Python applications.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=metadata["__author__"],
    author_email=metadata["__email__"],
    url="https://www.sqreen.com/",
    packages=find_packages(
        exclude=[
            "benchmarks.*", "benchmarks",
            "tests.*", "tests",
        ]
    ),
    package_dir={"sqreen": "sqreen"},
    include_package_data=True,
    install_requires=["py-mini-racer>=0.1.18", "sq-native>=0.6.1.0.1,<0.7.0"],
    license=metadata["__license__"],
    zip_safe=False,
    keywords="sqreen",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: Other/Proprietary License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Security",
    ],
    test_suite="tests",
    tests_require=[],
    entry_points={
        "console_scripts": ["sqreen-start = sqreen.bin.protect:protect"]
    },
)


class BuildExtFailed(Exception):
    pass


class optional_build_ext(build_ext):
    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildExtFailed

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError):
            raise BuildExtFailed


def run_setup(install_extensions=True):
    setup_options = {}
    if install_extensions:
        setup_options["cmdclass"] = {"build_ext": optional_build_ext}
        setup_options["ext_modules"] = [
            Extension("sqreen._vendors.wrapt._wrappers", ["sqreen/_vendors/wrapt/_wrappers.c"])
        ]
    setup_options.update(base_setup)
    setup(**setup_options)


SQREEN_INSTALL_EXTENSIONS = os.getenv("SQREEN_INSTALL_EXTENSIONS")
if SQREEN_INSTALL_EXTENSIONS is not None:
    # Build with or without extensions according to the environment variable.
    run_setup(install_extensions=SQREEN_INSTALL_EXTENSIONS.lower() in ("true", "1", "yes", "y"))
else:
    # By default, try to build the extensions and fallback to a pure-Python installation.
    try:
        run_setup(install_extensions=True)
    except BuildExtFailed:
        print("Install sqreen without extensions because they could not be compiled.")
        run_setup(install_extensions=False)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base webserver module
"""


class BaseWebServer(object):
    """Base class for webservers."""

    def various_infos(self):
        """Returns information sent during startup."""
        return {
            "name": "unknown",
        }

    @classmethod
    def detect(cls):
        """Returns an instance if the webserver was detected."""
        return cls()


class FallbackWebServer(BaseWebServer):
    """Default webserver if none was detected."""



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Condition evaluator."""

import operator
from collections import deque
from functools import wraps
from itertools import islice
from logging import getLogger

from .binding_accessor import BindingAccessor
from .exceptions import SqreenException
from .utils import (
    convert_to_string,
    is_unicode,
    iterkeys,
    itervalues,
    to_unicode_safe,
)

LOGGER = getLogger(__name__)


class ConditionError(SqreenException):
    """Base class for condition errors."""


class ConditionValueError(ConditionError):
    """Exception raised when the condition is invalid."""


class ConditionEvaluationError(ConditionError):
    """Exception raised at runtime when evaluation cannot complete."""


class ConditionRecursionError(ConditionError):
    """Exception raised when the condition is too deeply nested."""


def str_included(haystack, needle, min_needle_length=0):
    """Check if a needle is contained into the haystack.

    Haystack is an object and needle is a string. This function
    will do its best to convert the needle and the haystack to a
    string and might use str() on unknown objects, in this case,
    if needle is found it returns True, otherwise it raises a
    ConditionEvaluationError.
    """
    # Convert needle to unicode and check it is long enough
    needle = to_unicode_safe(needle)
    if needle is None or len(needle) < min_needle_length:
        return False

    if is_unicode(haystack):
        # Simplest case, needle and haystack are unicode, substring match
        return needle in haystack
    elif isinstance(haystack, (list, tuple)):
        # Iterate over all items, if any of the item contains needle, we return True
        # otherwise, if any of the items has an unexpected type, we try to convert it
        # to unicode and look for needle in it, otherwise we raise an error
        error = None
        for idx, elem in enumerate(haystack):
            if is_unicode(elem):
                # Both needle and elem are unicode, substring match
                if needle in elem:
                    return True
            else:
                try:
                    if needle in convert_to_string(elem):
                        return True
                except ValueError:
                    # Best effort convertion to unicode
                    elem_safe = to_unicode_safe(elem)
                    if elem_safe is not None and needle in elem_safe:
                        return True
                    # if nothing was found and type is unknown we have an error
                    if error is None and not isinstance(elem, bytes):
                        error = ConditionEvaluationError(
                            "unexpected input type {!r} in list (index {})".format(type(elem), idx))
        if error is not None:
            # Raise only the first error
            raise error
    else:
        try:
            if needle in convert_to_string(haystack):
                return True
        except ValueError:
            # Best effort, we check for needle in a repr of the haystack
            haystack_safe = to_unicode_safe(haystack)
            if haystack_safe is not None and needle in haystack_safe:
                return True
            if not isinstance(haystack, bytes):
                # raise if we don't find anything (we were not expecting this type at first)
                raise ConditionEvaluationError("unexpected input type {!r}".format(type(haystack)))

    return False


def hash_value_includes(value, iterable, min_value_size, max_iterations=1000):
    """Check whether a nested iterable is included into value object.

    The iterable can be a combination of dicts and lists. The argument
    min_value_size  is used to avoid comparison on small strings: For example,
    there is no possible SQL injection below 8 characters.
    """
    # Early stop.
    if iterable is None or value is None:
        return False

    iteration = 0
    remaining_iterables = deque([iterable], maxlen=max_iterations)
    while remaining_iterables:

        iteration += 1
        # If we have a very big or nested iterable, return True to execute the
        # rule.
        if iteration >= max_iterations:
            return True

        iterable = remaining_iterables.popleft()
        # If we get an iterable, add it to the list of remaining iterables.
        if isinstance(iterable, dict):
            remaining_iterables.extend(islice(itervalues(iterable), max_iterations))
        elif isinstance(iterable, (list, tuple)):
            remaining_iterables.extend(islice(iter(iterable), max_iterations))
        elif str_included(value, iterable, min_value_size):
            return True

    return False


def hash_key_includes(patterns, iterable, min_value_size, max_iterations=1000):
    """Check whether a nested iterable key matches a pattern or a list of patterns.

    The iterable can be a combination of dicts and lists. The argument
    min_value_size is used to avoid comparison on small strings: for example,
    there is no possible MongoDB injection below 1 characters.
    """
    # Early stop.
    if iterable is None or patterns is None:
        return False

    iteration = 0
    keys = deque(maxlen=max_iterations)
    remaining_iterables = deque([iterable], maxlen=max_iterations)
    while remaining_iterables:

        while keys:
            if str_included(patterns, keys.popleft(), min_value_size):
                return True

        iteration += 1
        # If we have a very big or nested iterable, return True to execute the
        # rule.
        if iteration >= max_iterations:
            return True

        iterable = remaining_iterables.popleft()
        # If we get an iterable, add it to the list of remaining iterables.
        if isinstance(iterable, dict):
            # We have new keys to test, remaining_iterables won't be empty
            keys.extend(islice(iterkeys(iterable), max_iterations))
            remaining_iterables.extend(islice(itervalues(iterable), max_iterations))
        elif isinstance(iterable, (list, tuple)):
            remaining_iterables.extend(islice(iter(iterable), max_iterations))

    return False


def unpack_parameters(f):
    """Unpack first argument into multiple arguments."""
    @wraps(f)
    def wrapper(values, **kwargs):
        return f(*values, **kwargs)
    return wrapper


def coerce_operator_types(operator):
    """Convert operator arguments when they are incompatibles.

    The items can be either integers, strings, bytes, list of items and dict of
    items. In order to avoid complexity, only the first level of items are
    converted. Operators are responsible to convert nested items.
    """
    @wraps(operator)
    def wrapper(*args, **kwargs):
        arg_types = [type(arg) for arg in args]
        # do nothing if all types are the same
        if arg_types and arg_types.count(arg_types[0]) != len(arg_types):
            # otherwise, convert the types to unicode whenever possible
            nargs = []
            for arg in args:
                if isinstance(arg, bytes):
                    arg = arg.decode('utf-8', errors='replace')
                elif isinstance(arg, (list, tuple)):
                    arg = [
                        (v.decode('utf-8', errors='replace') if isinstance(v, bytes) else v)
                        for v in arg
                    ]
                elif isinstance(arg, dict):
                    # Convert the key only (values are not used by basic operators like %include)
                    arg = {
                        (k.decode('utf-8', errors='replace') if isinstance(k, bytes) else k): v
                        for k, v in arg.items()
                    }
                nargs.append(arg)
            args = nargs
        return operator(*args, **kwargs)
    return wrapper


OPERATORS = {
    "%and": all,
    "%or": any,
    "%equals": unpack_parameters(coerce_operator_types(operator.eq)),
    "%not_equals": unpack_parameters(coerce_operator_types(operator.ne)),
    "%gt": unpack_parameters(coerce_operator_types(operator.gt)),
    "%gte": unpack_parameters(coerce_operator_types(operator.ge)),
    "%lt": unpack_parameters(coerce_operator_types(operator.lt)),
    "%lte": unpack_parameters(coerce_operator_types(operator.le)),
    "%include": unpack_parameters(coerce_operator_types(operator.contains)),
    "%hash_val_include": unpack_parameters(coerce_operator_types(hash_value_includes)),
    "%hash_key_include": unpack_parameters(coerce_operator_types(hash_key_includes)),
}

OPERATORS_ARITY = {
    "%equals": 2,
    "%not_equals": 2,
    "%gt": 2,
    "%gte": 2,
    "%lt": 2,
    "%lte": 2,
    "%include": 2,
    "%hash_val_include": 3,
    "%hash_key_include": 3,
}


def is_condition_empty(condition):
    """Return True if the condition is no-op, False otherwise."""
    if condition is None:
        return True
    elif isinstance(condition, bool):
        return False
    elif isinstance(condition, dict):
        return len(condition) == 0
    else:
        LOGGER.warning("Invalid precondition type: %r", condition)
        return True


def compile_condition(condition, level):
    """Compile a raw condition and validate it.

    Values are replaced by BindingAccessor instances and operator validity and
    arity are checked.
    """
    if level <= 0:
        raise ConditionRecursionError("compile recursion depth exceeded")

    if isinstance(condition, bool):
        return condition

    if not isinstance(condition, dict):
        raise ConditionValueError(
            "condition should be a dict, got {}".format(type(condition))
        )

    compiled = {}

    for _operator, values in condition.items():

        # Check operator validity.
        if _operator not in OPERATORS:
            raise ConditionValueError("unkown operator {!r}".format(_operator))

        # Check operator arity.
        arity = OPERATORS_ARITY.get(_operator, len(values))
        if len(values) != arity:
            raise ConditionValueError(
                "bad arity for operator {!r}: expected {}, got {}".format(
                    _operator, arity, len(values)
                )
            )

        # Check types.
        if not isinstance(values, list):
            raise ConditionValueError(
                "values should be an array, got {}".format(type(values))
            )

        compiled_values = []
        for value in values:
            if isinstance(value, bool):
                compiled_values.append(value)
            elif isinstance(value, dict):
                compiled_values.append(compile_condition(value, level - 1))
            else:
                # XXX we should warn on this because not all types will nicely convert to string
                # and most of the time it will repr() the value.
                compiled_values.append(BindingAccessor(to_unicode_safe(value)))

        compiled[_operator] = compiled_values

    return compiled


def evaluate(value, level=1, **kwargs):
    """Evaluate a value."""

    if isinstance(value, BindingAccessor):
        return value.resolve(**kwargs)

    elif isinstance(value, dict):

        if level <= 0:
            raise ConditionRecursionError("resolve recursion depth exceeded")

        for operator_name, values in value.items():
            operator_func = OPERATORS.get(operator_name)
            if operator_func is None:
                raise ConditionValueError("unkown operator {!r}".format(operator_name))

            # Create a lazily evaluated value generator
            values = (evaluate(v, level - 1, **kwargs) for v in values)
            result = operator_func(values)
            if result is False:
                return False

        return True

    return value


class ConditionEvaluator(object):
    """Evaluate a condition, resolving literals using BindingAccessor.

    {"%and": ["true", "true"]} -> true
    {"%or": ["true", "false"]} -> true
    {"%and": ["false", "true"]} -> false
    {"%equal": ["coucou", "#.args[0]"]} -> "coucou" == args[0]
    {"%hash_val_include": ["toto is a small guy", "#.request_params", 0]} ->
        true if one value of request params in included
        in the sentence 'toto is a small guy'.

    Combined expressions:
    { "%or":
        [
            {"%hash_val_include": ["AAA", "#.request_params", 0]},
            {"%hash_val_include": ["BBB", "#.request_params", 0]},
        ]
    }
    will return true if one of the request_params includes either AAA or BBB.
    """

    def __init__(self, condition):
        self.raw_condition = condition
        self.compiled = compile_condition(condition, 10)

    def evaluate(self, **kwargs):
        """Evaluate the compiled condition and return the result."""
        return evaluate(self.compiled, 10, **kwargs)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.raw_condition)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Package for webserver detection
"""
from .base import FallbackWebServer
from .uwsgi_webserver import UWSGIWebServer

_webserver_classes = [
    UWSGIWebServer
]


def get_webserver():
    """Detect the currently running webserver."""
    for cls in _webserver_classes:
        obj = cls.detect()
        if obj is not None:
            return obj
    return FallbackWebServer()


__all__ = [cls.__name__ for cls in _webserver_classes] \
    + ["FallbackWebServer", "get_webserver"]



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Sanitizer used to remove sensitive data from our payload"""

import json
import logging
import re

from . import config
from .utils import HAS_TYPING, flatten, is_string, is_unicode

if HAS_TYPING:
    from typing import (
        AbstractSet,
        Any,
        FrozenSet,
        Iterable,
        Iterator,
        Mapping,
        MutableMapping,
        Pattern,
        Set,
        Tuple,
    )
else:
    from collections import Iterable, Mapping


LOGGER = logging.getLogger(__name__)

MASK = '<Redacted by Sqreen>'


def compile_sensitive_regex():
    # type: () -> Pattern[str]
    try:
        pattern = config.CONFIG["STRIP_SENSITIVE_REGEX"]
        if not isinstance(pattern, str):
            raise TypeError
        return re.compile(pattern)
    except (TypeError, re.error):
        LOGGER.warning("Invalid regexp configuration: %r", config.CONFIG["STRIP_SENSITIVE_REGEX"])
        pattern = config.CONFIG_DEFAULT_VALUE["STRIP_SENSITIVE_REGEX"]
        if not isinstance(pattern, str):
            raise TypeError("Invalid default regex configuration")
        return re.compile(pattern)


SENSITIVE_KEYS = frozenset([k.strip() for k in config.CONFIG["STRIP_SENSITIVE_KEYS"].split(',')])  # type: FrozenSet[str]
SENSITIVE_VALUES = frozenset()  # type: FrozenSet[str]
SENSITIVE_REGEX = compile_sensitive_regex()
LOGGER.debug("Using sensitive keys %s", ", ".join(SENSITIVE_KEYS))
LOGGER.debug("Using sensitive regex %s", SENSITIVE_REGEX.pattern)


def sanitize(data, sensitive_keys=SENSITIVE_KEYS,
             sensitive_regex=SENSITIVE_REGEX,
             sensitive_values=SENSITIVE_VALUES):
    # type: (Any, AbstractSet[str], Pattern[str], AbstractSet[str]) -> Tuple[Any, AbstractSet[str]]
    """
    Sanitize sensitive data from an object. Return a 2-tuple with a sanitized
    copy of the data and a list of values that were sanitized.
    """
    sensitive_data = set()  # type: Set[str]

    if is_string(data):
        if not is_unicode(data):
            data = data.decode("utf-8", errors="replace")
        if data in sensitive_values or sensitive_regex.search(data):
            sensitive_data.add(data)
            data = MASK
        return data, sensitive_data

    elif isinstance(data, Mapping):
        new_dict = {}
        for k, v in data.items():
            if k in sensitive_keys:
                new_dict[k] = MASK
                if is_string(v):
                    sensitive_data.add(v)
                elif isinstance(v, Iterable):
                    keys, values = flatten(v)
                    sensitive_data.update(keys)
                    sensitive_data.update(values)
            else:
                ret_data, child_sensitive_data = sanitize(
                    v, sensitive_keys=sensitive_keys,
                    sensitive_regex=sensitive_regex,
                    sensitive_values=sensitive_values)
                new_dict[k] = ret_data
                sensitive_data.update(child_sensitive_data)
        return new_dict, sensitive_data

    elif isinstance(data, Iterable):
        new_list = []
        for v in data:
            ret_data, child_sensitive_data = sanitize(
                v, sensitive_keys=sensitive_keys,
                sensitive_regex=sensitive_regex,
                sensitive_values=sensitive_values)
            new_list.append(ret_data)
            sensitive_data.update(child_sensitive_data)
        return new_list, sensitive_data

    return data, sensitive_data


def sanitize_attacks(attacks, sensitive_values,
                     sensitive_regex=SENSITIVE_REGEX):
    # type: (Iterable[MutableMapping], AbstractSet[str], Pattern[str]) -> Iterator[MutableMapping]
    """
    Sanitize sensitive data from a list of attacks. Return the sanitized
    attacks.
    """
    for attack in attacks:
        infos = attack.get("infos")
        if infos is None:
            continue

        waf_data = infos.get("waf_data")
        if waf_data is not None:
            try:
                if isinstance(waf_data, bytes):
                    waf_data = waf_data.decode("utf-8", errors="replace")
                waf_data = json.loads(waf_data)
                assert isinstance(waf_data, list)
            except (UnicodeDecodeError, ValueError, AssertionError):
                waf_data = []

            normalized_sensitive_values = \
                {values.lower() for values in sensitive_values}
            new_waf_data = []
            for item in waf_data:
                filters = item.get("filter")
                if filters is not None:
                    for filter_item in filters:
                        resolved_value = filter_item.get("resolved_value")
                        if resolved_value is not None \
                                and resolved_value.lower() \
                                in normalized_sensitive_values:
                            filter_item["match_status"] = MASK
                            filter_item["resolved_value"] = MASK
                new_waf_data.append(item)

            infos["waf_data"] = json.dumps(new_waf_data, separators=(",", ":"))
        else:
            attack["infos"], _ = sanitize(
                infos, sensitive_keys=frozenset(),
                sensitive_regex=sensitive_regex,
                sensitive_values=sensitive_values)

        yield attack


def sanitize_exceptions(exceptions, sensitive_values):
    # type: (Iterable[Mapping], AbstractSet[str]) -> Iterator[Mapping]
    """
    Sanitize sensitive data from a list of exceptions. Return the sanitized
    exceptions.
    """
    for exc in exceptions:
        infos = exc.get("infos")
        if infos is not None:
            # We know the request contains PII, never send args
            # TODO more fine grained filtering
            args = infos.get("args")
            if args is not None and sensitive_values:
                infos.pop("args", None)

            waf_infos = infos.get("waf")
            if waf_infos is not None and sensitive_values:
                waf_infos.pop("args", None)

        yield exc



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Store a Django request and response in the storage and clear the request
context.
"""
from logging import getLogger

from ..frameworks.django_framework import DjangoRequest, DjangoResponse
from .record_request_context import RecordRequestContext

LOGGER = getLogger(__name__)


class RecordRequestContextDjango(RecordRequestContext):

    def pre(self, instance, args, kwargs, **options):
        self._store_request(DjangoRequest(args[0]))

    def post(self, instance, args, kwargs, result=None, **options):
        self.storage.store_response(DjangoResponse(result))
        self._clear_request()

    def failing(self, instance, args, kwargs, **options):
        """ Post is always called in a Django Middleware, don't clean the
        request right now as it may be needed in a post callback
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Record the current request in flask application
"""
from logging import getLogger

from ..frameworks.flask_framework import FlaskRequest, FlaskResponse
from .record_request import RecordRequest

LOGGER = getLogger(__name__)


class RecordRequestFlask(RecordRequest):
    def pre(self, instance, args, kwargs, **options):
        from flask import _request_ctx_stack

        reqctx = _request_ctx_stack.top
        if reqctx is not None:
            self._store_request(FlaskRequest(reqctx.request))

    def post(self, instance, args, kwargs, result=None, **options):
        self.storage.store_response(FlaskResponse(result))



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Contains the hook_point, the code actually executed in place of
application code
"""
import logging
import sys
from traceback import extract_stack

from .._vendors.wrapt import FunctionWrapper
from ..constants import ACTIONS, LIFECYCLE_METHODS, VALID_ACTIONS_PER_LIFECYCLE
from ..exceptions import (
    ActionBlock,
    ActionRedirect,
    AttackBlocked,
    MissingDataException,
    SqreenException,
)
from ..remote_exception import RemoteException
from ..runtime_storage import runtime
from ..utils import Iterable, Mapping
from .helpers import guard_call

LOGGER = logging.getLogger(__name__)


def _compute_result_action(current_result, next_result, valid_actions):
    """ Compute the next result_action based on current one and the callback
    result
    """
    current_status = current_result.get("status")
    next_status = next_result.get("status")
    # Check the validity of action per method
    if next_status not in valid_actions:
        # Ignore it
        return current_result

    if current_status not in valid_actions:
        return next_result

    current_result_order = valid_actions.index(current_result.get("status"))
    next_result_order = valid_actions.index(next_result.get("status"))

    # Return the old one only if it has more priority
    if current_result_order < next_result_order:
        return current_result
    else:
        return next_result


def valid_args(new_args):
    """ Check that new_args match the format ([], {}), else returns None
    """
    try:
        args, kwargs = new_args
    except (TypeError, ValueError):
        msg = "Invalid number of args or type, %s"
        LOGGER.debug(msg, new_args)
        return False

    # Check that new_args[0] is a valid type for passing to *
    if not isinstance(args, Iterable) or isinstance(args, Mapping):
        msg = "Invalid type for args[0]: %s"
        LOGGER.debug(msg, args)
        return False

    # Check that new_args[1] is a valid type for passing to **
    if not isinstance(kwargs, Mapping):
        msg = "Invalid type for args[1]: %s"
        LOGGER.debug(msg, kwargs)
        return False

    return True


def execute_callback(
    queue, callback, method, instance, result_action, args, kwargs,
    storage=runtime, override_budget=None, **options
):
    """ Execute a callback protected inside a broad try except.

    The queue argument is used to push a RemoteException when an exception is raised by the callback.
    The result_action argument is a dict where each callback pushes its results.
    """

    callback_method = getattr(callback, method)
    callback_rule_name = callback.rule_name
    LOGGER.debug(
        "Running %s callback %s of module '%s'",
        method,
        callback_method,
        callback_method.__module__,
    )

    if args is None:
        args = tuple()

    if kwargs is None:
        kwargs = {}

    if override_budget is not None and callback.collaborative:
        options["__sqreen_override_budget"] = override_budget

    if result_action:
        # Give access to the previous callbacks result action
        # For example, it allows a callback to know if the return value
        # will be overwritten or if the args have been changed
        options["result_action"] = result_action

    try:
        result = callback_method(instance, args, kwargs, **options)

        # First process of result

        if not result:
            return result_action, args, kwargs

        if not callback.block:
            LOGGER.debug("%s cannot block, ignoring return value", callback)
            return result_action, args, kwargs

        if not isinstance(result, dict):
            msg = "%s returns %s during %s method"
            LOGGER.warning(msg, callback, type(result), method)
            return result_action, args, kwargs

        # Set rule_name if not set
        if result.get("rule_name") is None:
            result["rule_name"] = callback_rule_name

        valid_actions = VALID_ACTIONS_PER_LIFECYCLE[method]
        result_action = _compute_result_action(
            result_action, result, valid_actions
        )

        # We need to process MODIFY_ARGS for next callbacks
        if result_action.get("status") == ACTIONS["MODIFY_ARGS"]:
            if valid_args(result_action["args"]):
                args, kwargs = result_action["args"]

        return result_action, args, kwargs

    except MissingDataException:
        raise

    except Exception as exception:
        LOGGER.exception(
            "An exception occurred while trying to execute %s", callback
        )

        exc_info = sys.exc_info()
        current_request = storage.get_current_request()
        stack = extract_stack()

        if current_request:
            callback.record_exception(exception, exc_info, stack)
        else:
            callback_exception_payload = callback.exception_infos()

            # Try to recover some infos from the exception if it's a SqreenException
            exception_infos = {}
            if isinstance(exception, SqreenException):
                exception_infos.update(exception.exception_infos())  # pylint: disable=no-member

            remote_exception = RemoteException.from_exc_info(
                exc_info,
                callback_payload=callback_exception_payload,
                exception_payload=exception_infos,
                stack=stack,
            )
            queue.put(remote_exception)

        return result_action, args, kwargs


def execute_callbacks(
    queue, callbacks, method, instance, args=None, kwargs=None,
    storage=runtime, override_budget=None, **options
):
    """ Execute a list of callbacks method (pre/post/fail), catch any exception
    that could happens.

    Aggregate the callbacks result (format {"status": COMMAND}), compute the
    final ACTION to execute and return it.
    For every ACTION, each callback is executed even if the first callback
    detected an attack except specified otherwise with the "immediate" key.

    The performance cap budget can be ignored by the caller by specifying an
    `override_budget` value though it is currently only used by the PHP daemon.
    """
    result_action = {}
    # Test if override_budget is already negative
    overtime = override_budget <= 0 if override_budget is not None else None
    total_duration = 0
    budget = None

    # Execute the callbacks
    for callback in callbacks:

        if callback.whitelisted:
            continue

        # Budget must be updated before executing the first callback
        if overtime is None and budget is None:
            budget = override_budget or callback.get_remaining_budget()
            overtime = budget <= 0 if budget is not None else False

        # Budget is exhausted, skip the callback if possible
        if overtime and callback.skippable:
            LOGGER.debug("Callback %s.%s was skipped after %0.2fms",
                         callback.rule_name, method, total_duration)
            callback.record_overtime(lifecycle=method)
            continue

        with storage.trace() as callback_trace:
            result = execute_callback(
                queue, callback, method, instance, result_action, args, kwargs,
                storage=storage, override_budget=budget, **options
            )

        result_action, args, kwargs = result

        total_duration += callback_trace.duration_ms
        if budget is not None:
            budget -= callback_trace.duration_ms
            overtime = budget <= 0
            if overtime:
                callback.record_overtime(lifecycle=method)

        if result_action.get("immediate", False):
            # Stop execution of other callbacks on this hook point if the result must be immediate.
            break

    # It is either empty dict if no override or the last one
    return result_action


def execute_pre_callbacks(
    key, strategy, instance, args=None, kwargs=None, storage=runtime,
    override_budget=None
):
    """ Execute pre_callbacks. Pre callbacks will receive these arguments:
    (instance, args, kwargs)
    """
    pre_callbacks = strategy.get_pre_callbacks(key)
    if pre_callbacks:
        return guard_call(
            key,
            execute_callbacks,
            strategy.queue,
            pre_callbacks,
            LIFECYCLE_METHODS["PRE"],
            instance,
            args,
            kwargs,
            storage=storage,
            override_budget=override_budget,
        )
    return {}


def execute_failing_callbacks(
    key, strategy, instance, exc_info, args=None, kwargs=None, storage=runtime,
    override_budget=None
):
    """ Execute failing_callbacks. Failing callbacks will receive these arguments:
    (instance, args, kwargs, exc_info=exc_info)
    """
    failing_callbacks = strategy.get_failing_callbacks(key)
    if failing_callbacks:
        return guard_call(
            key,
            execute_callbacks,
            strategy.queue,
            failing_callbacks,
            LIFECYCLE_METHODS["FAILING"],
            instance,
            args,
            kwargs,
            exc_info=exc_info,
            storage=storage,
            override_budget=override_budget,
        )
    return {}


def execute_post_callbacks(
    key, strategy, instance, result, args=None, kwargs=None, storage=runtime,
    override_budget=None
):
    """ Execute post_callbacks. Post callbacks will receive these arguments:
    (instance, args, kwargs, result=result)
    """
    post_callbacks = strategy.get_post_callbacks(key)
    if post_callbacks:
        return guard_call(
            key,
            execute_callbacks,
            strategy.queue,
            post_callbacks,
            LIFECYCLE_METHODS["POST"],
            instance,
            args,
            kwargs,
            result=result,
            storage=storage,
            override_budget=override_budget,
        )
    return {}


def hook_point(strategy, hook_name, hook_method, original, storage=runtime):
    """ Execute the original method and pre/post/failing callbacks
    If an exception happens, create a RemoteException, call
    callback.exception_infos for more debugging infos and send it via
    the provided queue.
    """

    def wrapper(wrapped, instance, args, kwargs):
        LOGGER.debug(
            "Checking before hook point of %s for %s/%s",
            strategy,
            hook_name,
            hook_method,
        )
        strategy.before_hook_point()
        key = (hook_name, hook_method)

        # Call pre callbacks
        action = execute_pre_callbacks(
            key,
            strategy,
            instance,
            args,
            kwargs,
            storage=storage
        )

        if action.get("status") == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            raise AttackBlocked(action.get("rule_name"))
        elif action.get("status") == ACTIONS["ACTION_BLOCK"]:
            LOGGER.debug(
                "Action %s blocked the request", action.get("action_id")
            )
            raise ActionBlock(action.get("action_id"))
        elif action.get("status") == ACTIONS["ACTION_REDIRECT"]:
            LOGGER.debug(
                "Action %s redirected the request to %r",
                action.get("action_id"),
                action["target_url"],
            )
            raise ActionRedirect(action.get("action_id"), action["target_url"])
        elif action.get("status") == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")
        elif action.get("status") == ACTIONS["MODIFY_ARGS"]:
            if valid_args(action["args"]):
                args, kwargs = action["args"]

        # Call original method
        retry = True
        while retry is True:
            try:
                retry = False
                # Try to call original function
                result = wrapped(*args, **kwargs)
            except Exception:
                # In case of error, call fail callbacks with exception infos
                exc_info = sys.exc_info()

                # Either raise an exception, set a return value or retry
                action = execute_failing_callbacks(
                    key, strategy, instance, exc_info, args, kwargs,
                    storage=storage
                )

                if action.get("status") == ACTIONS["RAISE"]:
                    LOGGER.debug(
                        "Callback %s detected an attack",
                        action.get("rule_name"),
                    )
                    raise AttackBlocked(action.get("rule_name"))
                elif action.get("status") == ACTIONS["RETRY"]:
                    retry = True
                elif action.get("status") == ACTIONS["OVERRIDE"]:
                    return action.get("new_return_value")

                # Be sure to raise if no retry or override
                if retry is False:
                    raise

        # Then call post callback in reverse order to simulate decorator
        # behavior
        action = execute_post_callbacks(
            key, strategy, instance, result, args, kwargs, storage=storage,
        )

        if action.get("status") == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            raise AttackBlocked(action.get("rule_name"))
        elif action.get("status") == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")

        # And return the original value
        return result

    return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Custom error page for Pyramid
"""
from .headers_insert import convert_to_str
from .sqreen_error_page import BaseSqreenErrorPage


class SqreenErrorPagePyramid(BaseSqreenErrorPage):

    def pre(self, instance, args, kwargs, **options):

        context = args[2]
        ret = self.handle_exception(context)

        if ret is not None:
            status_code, content, headers = ret

            from pyramid.response import Response  # type: ignore

            response = Response(content, status_code=status_code)

            for header_name, header_value in convert_to_str(headers.items()):
                response.headers[header_name] = header_value

            return {
                "status": "override",
                "new_return_value": response
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#


class AuthenticationFailed(Exception):
    """The authentication token or application name was rejected by the
    Sqreen Ingestion service."""


class DataIngestionFailed(Exception):
    """The recorded data was rejected by the Sqreen Ingestion service."""


class UnexpectedStatusCode(Exception):
    """Unexpected error from the Sqreen Ingestion service."""



# -*- coding: utf-8 -*-
#
#  Copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Large file support

.. deprecated:: 3.4

    The VARBLOCK format is NOT recommended for general use, has been deprecated since
    Python-RSA 3.4, and will be removed in a future release. It's vulnerable to a
    number of attacks:

    1. decrypt/encrypt_bigfile() does not implement `Authenticated encryption`_ nor
       uses MACs to verify messages before decrypting public key encrypted messages.

    2. decrypt/encrypt_bigfile() does not use hybrid encryption (it uses plain RSA)
       and has no method for chaining, so block reordering is possible.

    See `issue #19 on Github`_ for more information.

.. _Authenticated encryption: https://en.wikipedia.org/wiki/Authenticated_encryption
.. _issue #19 on Github: https://github.com/sybrenstuvel/python-rsa/issues/13


This module contains functions to:

    - break a file into smaller blocks, and encrypt them, and store the
      encrypted blocks in another file.

    - take such an encrypted files, decrypt its blocks, and reconstruct the
      original file.

The encrypted file format is as follows, where || denotes byte concatenation:

    FILE := VERSION || BLOCK || BLOCK ...

    BLOCK := LENGTH || DATA

    LENGTH := varint-encoded length of the subsequent data. Varint comes from
    Google Protobuf, and encodes an integer into a variable number of bytes.
    Each byte uses the 7 lowest bits to encode the value. The highest bit set
    to 1 indicates the next byte is also part of the varint. The last byte will
    have this bit set to 0.

This file format is called the VARBLOCK format, in line with the varint format
used to denote the block sizes.

"""

import warnings

from . import key, common, pkcs1, varblock
from ._compat import byte


def encrypt_bigfile(infile, outfile, pub_key):
    """Encrypts a file, writing it to 'outfile' in VARBLOCK format.

    .. deprecated:: 3.4
        This function was deprecated in Python-RSA version 3.4 due to security issues
        in the VARBLOCK format. See the documentation_ for more information.

    .. _documentation: https://stuvel.eu/python-rsa-doc/usage.html#working-with-big-files

    :param infile: file-like object to read the cleartext from
    :param outfile: file-like object to write the crypto in VARBLOCK format to
    :param pub_key: :py:class:`rsa.PublicKey` to encrypt with

    """

    warnings.warn("The 'rsa.bigfile.encrypt_bigfile' function was deprecated in Python-RSA version "
                  "3.4 due to security issues in the VARBLOCK format. See "
                  "https://stuvel.eu/python-rsa-doc/usage.html#working-with-big-files "
                  "for more information.",
                  DeprecationWarning, stacklevel=2)

    if not isinstance(pub_key, key.PublicKey):
        raise TypeError('Public key required, but got %r' % pub_key)

    key_bytes = common.bit_size(pub_key.n) // 8
    blocksize = key_bytes - 11  # keep space for PKCS#1 padding

    # Write the version number to the VARBLOCK file
    outfile.write(byte(varblock.VARBLOCK_VERSION))

    # Encrypt and write each block
    for block in varblock.yield_fixedblocks(infile, blocksize):
        crypto = pkcs1.encrypt(block, pub_key)

        varblock.write_varint(outfile, len(crypto))
        outfile.write(crypto)


def decrypt_bigfile(infile, outfile, priv_key):
    """Decrypts an encrypted VARBLOCK file, writing it to 'outfile'

    .. deprecated:: 3.4
        This function was deprecated in Python-RSA version 3.4 due to security issues
        in the VARBLOCK format. See the documentation_ for more information.

    .. _documentation: https://stuvel.eu/python-rsa-doc/usage.html#working-with-big-files

    :param infile: file-like object to read the crypto in VARBLOCK format from
    :param outfile: file-like object to write the cleartext to
    :param priv_key: :py:class:`rsa.PrivateKey` to decrypt with

    """

    warnings.warn("The 'rsa.bigfile.decrypt_bigfile' function was deprecated in Python-RSA version "
                  "3.4 due to security issues in the VARBLOCK format. See "
                  "https://stuvel.eu/python-rsa-doc/usage.html#working-with-big-files "
                  "for more information.",
                  DeprecationWarning, stacklevel=2)

    if not isinstance(priv_key, key.PrivateKey):
        raise TypeError('Private key required, but got %r' % priv_key)

    for block in varblock.yield_varblocks(infile):
        cleartext = pkcs1.decrypt(block, priv_key)
        outfile.write(cleartext)


__all__ = ['encrypt_bigfile', 'decrypt_bigfile']



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Simple delivery method that directly call session on event
"""

from ..events import Attack, RequestRecord
from ..remote_exception import RemoteException


class SimpleDeliverer(object):
    """ Class responsible for send events to backend depending
    on their types
    """

    batch_size = 0
    original_max_staleness = 0
    max_staleness = 0

    def __init__(self, session):
        self.session = session

    def post_event(self, event):
        """ Post a single event
        """
        if isinstance(event, RemoteException):
            return self.session.post_sqreen_exception(event.to_dict())
        if isinstance(event, Attack):
            return self.session.post_attack(event.to_dict())
        if isinstance(event, RequestRecord):
            return self.session.post_request_record(event.to_dict())
        else:
            err_msg = "Unknown event type {}".format(type(event))
            raise NotImplementedError(err_msg)

    def post_metrics(self, metrics):
        """ Post metrics
        """
        metrics = [
            {k: v for k, v in metric.items() if k in ("name", "observation", "start", "finish")}
            for metric in metrics
        ]
        return self.session.post_metrics(metrics)

    def drain(self, resiliently):
        """ Since everything is posted at once nothing needs to be done here
        """
        pass

    def tick(self):
        """ Since everything is posted at once nothing needs to be done here
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Flask hook strategy
"""

from logging import getLogger

from ..middlewares.flask_middleware import FlaskMiddleware
from .framework import FrameworkStrategy

LOGGER = getLogger(__name__)


def replace_wsgi_app(original, middleware):
    return middleware(original)


class FlaskStrategy(FrameworkStrategy):
    """ Strategy for Flask peripheric callbacks.

    It injects functions that calls pre and post callbacks in the Flask
    request workflow
    """

    MODULE_NAME = "flask.app"
    HOOK_CLASS = "Flask"
    HOOK_METHOD = "full_dispatch_request"

    def __init__(
        self,
        strategy_key,
        observation_queue,
        queue,
        import_hook,
        before_hook_point=None,
    ):
        super(FlaskStrategy, self).__init__(
            strategy_key,
            observation_queue,
            queue,
            import_hook,
            before_hook_point,
        )

        self.middleware = FlaskMiddleware(self, observation_queue, queue)
        self.wrapper = replace_wsgi_app
